#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Schedule and timetabling interfaces
"""

import pytz

import zope.schema
from zope.annotation.interfaces import IAttributeAnnotatable
from zope.interface import Interface, Attribute
from zope.container.constraints import contains
from zope.container.interfaces import IContainer, IOrderedContainer
from zope.container.interfaces import IContained

from schooltool.app.interfaces import ISchoolToolCalendarEvent
from schooltool.app.interfaces import ISchoolToolCalendar
from schooltool.app.interfaces import IHaveCalendar
from schooltool.app.utils import vocabulary
from schooltool.calendar.interfaces import ICalendar

from schooltool.common import SchoolToolMessage as _


#
#  Schedule
#

activity_types = vocabulary(
    [("lesson", _("Lesson")),
     ("homeroom", _("Homeroom")),
     ("free", _("Free")),
     ("lunch", _("Lunch")),
     ])


class IPeriod(IContained):
    """A period of activity."""

    title = zope.schema.TextLine(
        title=u"Title of the period.",
        required=False)

    activity_type = zope.schema.Choice(
        title=_("Activity type."),
        required=True,
        default="lesson",
        vocabulary=activity_types)


class IMeeting(Interface):
    """A meeting represents a lesson or other scheduled activity."""

    dtstart = zope.schema.Datetime(
        title=u"Time of the start of the event",
        required=True)

    duration = zope.schema.Timedelta(
        title=u"Timedelta of the duration of the event",
        required=True)

    period = zope.schema.Object(
        title=u"The period to schedule.",
        schema=IPeriod,
        required=False)

    meeting_id = zope.schema.TextLine(
        title=u"Unique identifier of a meeting (lesson)",
        description=u"""
        The meeting_id is an arbitrary identifier of a meeting (lesson),
        unique within the schedule that generated it.
        """,
        required=False)

    def clone(dtstart=None, duration=None,
              period=None, meeting_id=None):
        """Return a copy of the meeting with replaced given values."""


class ISchedule(Interface):
    """A schedule of meetings."""

    title = zope.schema.TextLine(
        title=u"Title of the timetable",
        required=True)

    first = zope.schema.Date(
        title=_("First scheduled day"),
        required=True)

    last = zope.schema.Date(
        title=_("Last scheduled day"),
        required=True)

    timezone = zope.schema.Choice(
        title=u"Time Zone",
        description=u"Meetings time zone",
        values=pytz.common_timezones,
        required=True)

    def iterMeetings(date, until_date=None):
        """Yields meetings for the given date range."""


class IMeetingException(IMeeting):
    """A persistent exception meeting."""


class IScheduleExceptions(Interface):

    # XXX: if exception values were objects instead of meeting lists
    #      we could log additional info, like reason for the exception
    exceptions = zope.schema.Dict(
        title=_("Exceptions"),
        description=_("Exceptions by date."),
        key_type=zope.schema.Date(
            title=_("Exception date")),
        value_type=zope.schema.List(
            title=_("Meetings"),
            value_type = zope.schema.Object(
                title=_("Meeting"),
                schema=IMeetingException),
            ),
        )

    def iterOriginalMeetings(date, until_date=None):
        """Yields meetings disregarding exception dates."""


class IScheduleWithExceptions(ISchedule, IScheduleExceptions):
    """Schedule with exception days."""


class IScheduleContainer(IContainer, IScheduleWithExceptions):
    """A container of schedules.

    The container itself is as a big composite schedule.
    """
    contains(ISchedule)

    first = zope.schema.Date(
        title=u"First scheduled day",
        required=False)

    last = zope.schema.Date(
        title=u"Last scheduled day",
        required=False)


#
#  Day templates
#

class IDayTemplate(IOrderedContainer):
    title = zope.schema.TextLine(
        title=u"Title of the day",
        required=False)


class IDayTemplateContainer(IOrderedContainer):
    """Ordered container of day templates."""
    contains(IDayTemplate)


class ITimeSlot(IContained):
    """Time slot designated for an activity."""

    tstart = zope.schema.Time(
        title=u"Time of the start of the event",
        required=True)

    duration = zope.schema.Timedelta(
        title=u"Timedelta of the duration of the event",
        required=True)

    activity_type = zope.schema.Choice(
        title=_("Activity type"),
        required=False,
        default="lesson",
        vocabulary=activity_types)


class IDayTemplateSchedule(IContained):
    """Day templates scheduled by date."""

    templates = zope.schema.Object(
        title=u"The template container.",
        schema=IDayTemplateContainer,
        required=True)

    def iterDates(dates):
        """Yield day templates for given dates."""


class ICalendarDayTemplates(IDayTemplateSchedule):
    """Day templates."""

    starting_index = zope.schema.Int(
        title=u"Starting date should start as Nth day",
        default=0,
        required=True)

    def getDay(schedule, date):
        """Get template for the given date."""


class IWeekDayTemplates(IDayTemplateSchedule):
    """Iterator of day templates."""

    def getWeekDayKey(weekday):
        """Get weekday template container key."""

    def getWeekDay(weekday):
        """Get weekday template."""


class ISchooldays(Interface):
    """XXX: should be moved to term, as schooldays are defined there."""

    def iterDates(dates):
        """Iterate dates that are schooldays."""

    def __iter__():
        """Yield all schoolday dates."""

    def __contains__(date):
        """Return whether the date is a schoolday."""


class ISchoolDayTemplates(IDayTemplateSchedule):
    """Iterator that rotates on schooldays (as opposed to rotating on
    calendar days)"""

    starting_index = zope.schema.Int(
        title=u"Starting date should start as Nth day",
        default=0,
        required=True)


#
#  Timetabling
#


class ITimetable(IScheduleWithExceptions):
    """The schedule of meetings built from day templates."""

    periods = zope.schema.Object(
        title=u"Periods",
        schema=IDayTemplateSchedule,
        required=True)

    time_slots = zope.schema.Object(
        title=u"Time slots",
        schema=IDayTemplateSchedule,
        required=True)


class ITimetableContainerBase(Interface):

    default = zope.schema.Object(
        title=u"The default timetable",
        schema=ITimetable,
        required=False)


class ITimetableContainer(IContainer, ITimetableContainerBase):
    """A container of timetables."""
    contains(ITimetable)


class ISelectedPeriodsScheduleRead(ISchedule):

    timetable = zope.schema.Object(
        title=u"Timetable to filter meetings from",
        schema=ITimetable,
        required=False)

    periods = Attribute(
        """Iterate only over meetings for these periods.""")

    consecutive_periods_as_one = zope.schema.Bool(
        title=_("Treat consecutive periods as one meeting"),
        default=False,
        required=False)

    def hasPeriod(period):
        """Is the period added to this schedule."""


class ISelectedPeriodsScheduleWrite(Interface):

    def addPeriod(period):
        """Schedule meetings for this period."""

    def removePeriod(period):
        """Unschedule meeting for this period."""


class ISelectedPeriodsSchedule(ISelectedPeriodsScheduleRead,
                               ISelectedPeriodsScheduleWrite):
    """Schedule composed of meetings from another schedule with
    selected periods only."""

    def addPeriod(period):
        """Schedule meetings for this period."""

    def removePeriod(period):
        """Unschedule meeting for this period."""


#
#  Calendar
#


class IScheduleCalendarEvent(IMeeting, ISchoolToolCalendarEvent):
    """Calendar event that forms a schedule."""

    schedule = zope.schema.Object(
        title=u"Schedule that generated this event.",
        schema=ISchedule,
        required=False)


class IImmutableScheduleCalendar(ICalendar):
    """Calendar representation of a schedule."""


class IScheduleCalendar(ISchoolToolCalendar):
    """Persistent calendar of a schedule."""

    def updateSchedule(schedule):
        """Update calendar with events from this schedule."""

    def removeSchedule(schedule):
        """Remove events generated by this schedule."""


#
#  Integration
#


class IHaveSchedule(IAttributeAnnotatable, IHaveCalendar):
    """Marker interface for objects that schedule existing timetables."""


class IHaveTimetables(IAttributeAnnotatable):
    """Marker interface for objects that have timetables."""


#
#  Security
#


class IScheduleParentCrowd(Interface):
    """A crowd object that is used on a schedule's parent.

    This is just a marker interface.
    """

    def contains(principal):
        """Return True if principal is in the crowd."""



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.app.main.
"""

import os
import sys
import unittest
import doctest

from zope.component import getUtility
from zope.testing import cleanup
from zope.traversing.interfaces import IContainmentRoot
from zope.interface import directlyProvides

here = os.path.dirname(__file__)
ftesting_zcml = os.path.join(here, '..', 'ftesting.zcml')


def doctest_main():
    """Tests for main().

    Main does nothing more but configures SchoolTool, prints the startup time,
    and starts the main loop.

    Since we don't want to actually create disk files and start a web server in
    a test, we will set up some stubs.

        >>> from schooltool.app.main import Options
        >>> options = Options()
        >>> class ConfigStub:
        ...     pass
        >>> options.config = ConfigStub()

        >>> def load_options_stub(argv):
        ...     return options
        >>> def setup_stub(opts):
        ...     print "Performing setup..."
        ...     assert opts is options
        >>> from schooltool.app.main import SchoolToolServer
        >>> server = SchoolToolServer()
        >>> server.load_options = load_options_stub
        >>> server.setup = setup_stub

    Now we will run main().

        >>> server.main(['sb.py'])
        Performing setup...

    """


def doctest_load_options():
    """Tests for load_options().

        >>> from zope.app.testing import setup
        >>> setup.placelessSetUp()

    We will use a sample configuration file that comes with these tests.

        >>> import os
        >>> from schooltool.app import tests
        >>> test_dir = os.path.dirname(tests.__file__)
        >>> sample_config_file = os.path.join(test_dir, 'sample.conf')
        >>> empty_config_file = os.path.join(test_dir, 'empty.conf')

    load_options will report errors to stderr.  We need to temporarily
    redirect stderr to stdout, because otherwise doctests will not see the
    output.

        >>> old_stderr = sys.stderr
        >>> sys.stderr = sys.stdout

    Load options parses command line arguments and the configuration file.

        >>> from schooltool.app.main import SchoolToolServer
        >>> server = SchoolToolServer()
        >>> o = server.load_options(['st.py', '-c', sample_config_file])
        Reading configuration from ...sample.conf

    Some options come from the command line

        >>> o.config_file
        '...sample.conf'
        >>> o.pack
        False

    Some come from the config file

        >>> o.config.error_log_file
        ['/var/log/schooltool/schooltool.log', 'STDERR']
        >>> o.config.web_access_log_file
        ['/var/log/schooltool/web-access.log']
        >>> o.config.reportlab_fontdir
        '/fonts/liberation:/fonts/ubuntu'

    `load_options` can also give you a nice help message and exit with status
    code 0.

        >>> try:
        ...     o = server.load_options(['st.py', '-h'])
        ... except SystemExit, e:
        ...     print '[exited with status %s]' % e
        Usage: st.py [options]
        Options:
          -c, --config xxx       use this configuration file
          -h, --help             show this help message
          --pack                 pack the database
          -r, --restore-manager password
                                 restore the manager user with the provided password
                                 (read password from the standart input if 'password'
                                 is '-')
        [exited with status 0]

    Here's what happens, when you use an unknown command line option.

        >>> try:
        ...     o = server.load_options(['st.py', '-q'])
        ... except SystemExit, e:
        ...     print '[exited with status %s]' % e
        st.py: option -q not recognized
        Run st.py -h for help.
        [exited with status 1]

    Here's what happens when the configuration file cannot be found

        >>> try:
        ...     o = server.load_options(['st.py', '-c', 'nosuchfile'])
        ... except SystemExit, e:
        ...     print '[exited with status %s]' % e
        Reading configuration from nosuchfile
        st.py: error opening file ...nosuchfile: ...
        [exited with status 1]

    Here's what happens if you do not specify a storage section in the
    configuration file.

        >>> try:
        ...     o = server.load_options(['st.py', '-c', empty_config_file])
        ... except SystemExit, e:
        ...     print '[exited with status %s]' % e
        Reading configuration from ...empty.conf
        st.py: No storage defined in the configuration file.
        [exited with status 1]

    Cleaning up.

        >>> sys.stderr = old_stderr

        >>> setup.placelessTearDown()

    """


def doctest_configureReportlab():
    """Tests for configureReportlab.

        >>> from schooltool.app import pdf
        >>> from schooltool.app.main import SchoolToolServer

        >>> server = SchoolToolServer()

        >>> def setupStub(fontdir):
        ...     print 'reportlab set up: %s' % fontdir
        >>> realSetup = pdf.setUpFonts
        >>> pdf.setUpFonts = setupStub

        >>> old_stderr = sys.stderr
        >>> sys.stderr = sys.stdout

    First, if a null path is given, nothing happens (PDF support is
    left disabled):

        >>> server.configureReportlab(None)

    Now test the check that the font path is a directory:

        >>> server.configureReportlab(__file__)
        Warning: font directories '...test_main...' do not exist.
        PDF support disabled.

    We will cheat and temporarily override pdf.font_map:

        >>> pseudo_fontdir = os.path.dirname(__file__)
        >>> real_font_map = pdf.font_map
        >>> pdf.font_map = {'1': 'test_main.py', '2': 'nonexistent_file'}

        >>> server.configureReportlab(pseudo_fontdir)
        Warning: font '...nonexistent_file' does not exist...
        PDF support disabled.

    Now let's simulate a successful scenario:

        >>> del pdf.font_map['2']
        >>> server.configureReportlab(pseudo_fontdir)
        reportlab set up: ...tests...

    Cleaning up.

        >>> pdf.font_map = real_font_map
        >>> sys.stderr = old_stderr
        >>> pdf.setUpFonts = realSetup

    """


def doctest_setLanguage():
    """Tests for setLanguage.

        >>> from zope.app.testing import setup
        >>> setup.placelessSetUp()

    First, the 'automatic mode':

        >>> from schooltool.app.main import setLanguage
        >>> setLanguage('auto')

    The language adapter shouldn't have been installed:

        >>> from zope.i18n.interfaces import IUserPreferredLanguages
        >>> from zope.publisher.browser import TestRequest
        >>> request = TestRequest()
        >>> IUserPreferredLanguages(request).getPreferredLanguages()
        []

    Now, if we specify a language, a language adapter should be set up:

        >>> setLanguage('lt')
        >>> IUserPreferredLanguages(request).getPreferredLanguages()
        ('lt',)

    If the language list contains more than one language, you get a
    cookie language selector:

        >>> from schooltool.app.interfaces import ICookieLanguageSelector
        >>> setLanguage('en, lt')
        >>> upl = IUserPreferredLanguages(request)
        >>> ICookieLanguageSelector.providedBy(upl)
        True

        >>> upl.getLanguageList()
        ['en', 'lt']

        >>> upl.getPreferredLanguages()
        ('en',)

    We're done.

        >>> setup.placelessTearDown()

    """


def doctest_setup():
    """Tests for setup()

        >>> cleanup.setUp()

    setup() does everything except enter the main application loop:

    - sets up loggers
    - configures Zope 3 components
    - opens the database (optionally packs it)

    It is difficult to unit test, but we'll try.

        >>> from schooltool.app.main import Options, SchoolToolServer
        >>> from ZODB.MappingStorage import MappingStorage
        >>> from ZODB.DB import DB
        >>> options = Options()
        >>> class DatabaseConfigStub:
        ...     def open(self):
        ...         return DB(MappingStorage())
        >>> class ConfigStub:
        ...     thread_pool_size = 1
        ...     database = DatabaseConfigStub()
        ...     pid_file = ''
        ...     error_log_file = ['STDERR']
        ...     web_access_log_file = ['STDOUT']
        ...     attendance_log_file = ['STDOUT']
        ...     lang = 'lt'
        ...     reportlab_fontdir = ''
        ...     devmode = False
        ...     site_definition = ftesting_zcml
        >>> options.config = ConfigStub()

    Workaround to fix a Windows failure:

        >>> import logging
        >>> del logging.getLogger(None).handlers[:]

    And go!

        >>> server = SchoolToolServer()
        >>> db = server.setup(options)
        >>> print db
        <ZODB.DB.DB object at ...>

    The root object is SchoolToolApplication:

        >>> connection = db.open()
        >>> root = connection.root()
        >>> from zope.app.publication.zopepublication import ZopePublication
        >>> app = root.get(ZopePublication.root_name)
        >>> app
        <schooltool.app.app.SchoolToolApplication object at ...>

    The manager is a SchoolTool person:

        >>> from schooltool.person.interfaces import IPerson
        >>> IPerson.providedBy(app['persons']['manager'])
        True

    A web access logger has been set up:

        >>> logger1 = logging.getLogger('accesslog')
        >>> logger1.propagate
        False
        >>> logger1.handlers
        [<logging.StreamHandler ...>]

    A generic access logger has been set up too:

        >>> logger2 = logging.getLogger(None)
        >>> logger2.handlers
        [<logging.StreamHandler ...>]

    The language adapter shouldn't have been installed:

        >>> from zope.i18n.interfaces import IUserPreferredLanguages
        >>> from zope.publisher.browser import TestRequest
        >>> request = TestRequest()
        >>> IUserPreferredLanguages(request).getPreferredLanguages()
        ('lt',)

    ZODB.lock_file has been shut up:

        >>> logging.getLogger('ZODB.lock_file').disabled
        True

    We better clean up logging before we leave:

        >>> logging.getLogger('ZODB.lock_file').disabled = False
        >>> for logger in [logger1, logger2]:
        ...     del logger.handlers[:]
        ...     logger.propagate = True
        ...     logger.disabled = False

        >>> for logger in [logger1]:
        ...     logger.setLevel(0)

        >>> cleanup.tearDown()

    """



class ConfigStub(object):

    devmode = False
    site_definition = None


class OptionsStub(object):

    config = ConfigStub()
    config_file = ''


def doctest_bootstrapSchoolTool():
    r"""Tests for bootstrapSchoolTool()

        >>> cleanup.setUp()

    Normally, bootstrapSchoolTool is called when Zope 3 is fully configured

        >>> from schooltool.app.main import SchoolToolServer
        >>> server = SchoolToolServer()
        >>> options = OptionsStub()
        >>> options.config.site_definition = ftesting_zcml
        >>> server.configureComponents(options)

    When we start with an empty database, bootstrapSchoolTool creates a
    SchoolTool application in it.

        >>> import transaction
        >>> from ZODB.DB import DB
        >>> from ZODB.MappingStorage import MappingStorage
        >>> db = DB(MappingStorage())

        >>> server.bootstrapSchoolTool(db)

    Let's take a look...

        >>> connection = db.open()
        >>> root = connection.root()
        >>> from zope.app.publication.zopepublication import ZopePublication
        >>> app = root.get(ZopePublication.root_name)
        >>> app
        <schooltool.app.app.SchoolToolApplication object at ...>

    This new application object is the containment root

        >>> IContainmentRoot.providedBy(app)
        True

    It is also a site

        >>> from zope.component.interfaces import ISite
        >>> ISite.providedBy(app)
        True

    It has a local authentication utility

        >>> from zope.authentication.interfaces import IAuthentication
        >>> getUtility(IAuthentication, context=app)
        <schooltool.app.security.SchoolToolAuthenticationUtility object at ...>

    It has an initial user (username 'manager', password 'schooltool')

        >>> manager = app['persons']['manager']
        >>> manager.checkPassword('schooltool')
        True

    bootstrapSchoolTool doesn't do anything if it finds the root object already
    present in the database.

        >>> from schooltool.person.person import Person
        >>> manager = app['persons']['user1'] = Person('user1')
        >>> transaction.commit()
        >>> connection.close()

        >>> server.bootstrapSchoolTool(db)

        >>> connection = db.open()
        >>> root = connection.root()
        >>> 'user1' in root[ZopePublication.root_name]['persons']
        True

    Clean up

        >>> transaction.abort()
        >>> connection.close()
        >>> cleanup.tearDown()
    """


def doctest_restoreManagerUser():
    r"""Unit test for SchoolToolServer.restoreManagerUser

        >>> cleanup.setUp()

    We need a configured server:

        >>> from schooltool.app.main import SchoolToolServer
        >>> server = SchoolToolServer()
        >>> options = OptionsStub()
        >>> options.config.site_definition = ftesting_zcml
        >>> server.configureComponents(options)

    We also need an application (we are doing the full set up in here
    because else person factory local utility is not being
    registered):

        >>> import transaction
        >>> from zope.lifecycleevent import ObjectAddedEvent
        >>> from schooltool.app.interfaces import ApplicationInitializationEvent
        >>> from zope.event import notify
        >>> from zope.app.publication.zopepublication import ZopePublication
        >>> from ZODB.DB import DB
        >>> from ZODB.MappingStorage import MappingStorage

        >>> db = DB(MappingStorage())
        >>> connection = db.open()
        >>> root = connection.root()

        >>> from schooltool.app.app import SchoolToolApplication
        >>> app = SchoolToolApplication()
        >>> directlyProvides(app, IContainmentRoot)
        >>> root[ZopePublication.root_name] = app

        >>> save_point = transaction.savepoint(optimistic=True)
        >>> notify(ApplicationInitializationEvent(app))
        >>> notify(ObjectAddedEvent(app))

    Initially, there's no manager user in the database:

        >>> app['persons']['manager']
        Traceback (most recent call last):
          ...
        KeyError: 'manager'

    When we call restoreManagerUser, it gets created:

        >>> server.restoreManagerUser(app, 'schooltool')

        >>> manager = app['persons']['manager']
        >>> manager.checkPassword('schooltool')
        True

    Manager is by default a super user:

        >>> app['persons'].super_user is manager
        True

    To prevent this user from being deleted we add a dependency

        >>> from zope.app.dependable.interfaces import IDependable
        >>> IDependable(manager).dependents()
        (u'/persons/',)

    Let's break the manager user by forgetting his password:

        >>> marker = object()
        >>> manager.calendar = marker
        >>> manager.setPassword('randomrandom')

        >>> manager.checkPassword('schooltool')
        False

    If we call restoreManagerUser again, the object remains the same,
    but its password gets reset:

        >>> server.restoreManagerUser(app, 'schooltool')

        >>> manager = app['persons']['manager']
        >>> manager.checkPassword('schooltool')
        True

    The manager is the same:

        >>> manager.calendar is marker
        True

    Cleanup:

        >>> transaction.abort()
        >>> connection.close()
        >>> cleanup.tearDown()
    """


def test_setUpLogger():
    r"""Tests for setUpLogger.

    setUpLogger sets up a logger:

        >>> import logging
        >>> from schooltool.app.main import setUpLogger
        >>> setUpLogger('schooltool.just_testing',
        ...             ['STDERR', '_just_testing.log'],
        ...             '%(asctime)s %(message)s')

        >>> logger = logging.getLogger('schooltool.just_testing')
        >>> logger.propagate
        False
        >>> logger.handlers
        [<logging.StreamHandler ...>, <...UnicodeFileHandler ...>]
        >>> logger.handlers[0].stream is sys.stderr
        True
        >>> logger.handlers[0].formatter
        <logging.Formatter ...>
        >>> logger.handlers[0].formatter._fmt
        '%(asctime)s %(message)s'

    Let's clean up after ourselves (logging is messy):

        >>> del logger.handlers[:]
        >>> logger.propagate = True
        >>> logger.disabled = False
        >>> logger.setLevel(0)

        >>> import os
        >>> os.unlink('_just_testing.log')

    """



def doctest_CyclicPluginActionOrderException():
    """Tests for CyclicPluginActionOrderException.

    This exception converts a list of PluginDependency items to a human
    readable format.

       >>> from schooltool.app.main import PluginDependency
       >>> from schooltool.app.main import CyclicPluginActionOrderException

       >>> class PluginStub(object):
       ...     def __init__(self, name):
       ...         self.name = name
       ...     def __repr__(self):
       ...         return '<%s>' % self.name

       >>> dependencies = [
       ...     PluginDependency(PluginStub('alpha'), 'A',
       ...                      PluginStub('beta'), 'B'),
       ...     PluginDependency(PluginStub('gamma'), 'G',
       ...                      PluginStub('delta'), 'D',
       ...                      inverse=True),
       ...     PluginDependency(PluginStub('omega'), 'O')]

       >>> exception = CyclicPluginActionOrderException(dependencies)

       >>> print exception
       Cannot resolve plugin action order:
       <alpha> named "A" must be executed _before_ <beta> named "B".
       <gamma> named "G" must be executed _after_ <delta> named "D".
       <omega> named "O" must be executed _before_ another.

    """


def doctest_PluginActionSorter():
    """Tests for PluginActionSorter.

    The purpose of PluginActionSorter is to resolve the order in which
    the plugin actions should be executed.

        >>> from schooltool.app.main import PluginActionSorter

        >>> class ActionStub(object):
        ...     def __init__(self, exec_after, name, exec_before):
        ...         self.name = name
        ...         self.after = exec_after
        ...         self.before = exec_before
        ...     def __repr__(self):
        ...         return '<action %s>' % self.name

    The sorter takes a list of (adapter_name, factory) tuples that can be
    obtained with getAdapters.

        >>> actions = [
        ...     ActionStub([], 'C', []),
        ...     ActionStub([], 'A', ['B']),
        ...     ActionStub([], 'B', []),
        ...     ActionStub(['A'], 'Q', []),
        ...     ]

        >>> sorter = PluginActionSorter([(a.name, a) for a in actions])

    The sorter tries to maintain the original order of plugins if possible.

        >>> print sorter()
        [<action C>, <action A>, <action B>, <action Q>]

    When the order cannot be maintained, plugins are sorted by their
    execution dependencies.

        >>> actions = [
        ...     ActionStub([], 'A', []),
        ...     ActionStub([], 'B', []),
        ...     ActionStub(['D'], 'C', []),
        ...     ActionStub(['B'], 'D', ['A']),
        ...     ]

        >>> sorter = PluginActionSorter([(a.name, a) for a in actions])
        >>> print sorter()
        [<action B>, <action D>, <action A>, <action C>]

    In case of invalid dependencies, an exception is thrown.

        >>> actions = [
        ...     ActionStub([], 'A', []),
        ...     ActionStub(['A'], 'B', ['A']),
        ...     ]
        >>> sorter = PluginActionSorter([(a.name, a) for a in actions])
        >>> print sorter()
        Traceback (most recent call last):
        ...
        CyclicPluginActionOrderException: Cannot resolve plugin action order:
        <action A> named "A" must be executed _before_ <action B> named "B".
        <action A> named "A" must be executed _after_ <action B> named "B".

    The sorter also detects dependency cycles.

        >>> actions = [
        ...     ActionStub([], 'A', ['B']),
        ...     ActionStub([], 'B', []),
        ...     ActionStub(['B'], 'C', ['A']),
        ...     ]

        >>> sorter = PluginActionSorter([(a.name, a) for a in actions])
        >>> print sorter()
        Traceback (most recent call last):
        ...
        CyclicPluginActionOrderException: Cannot resolve plugin action order:
        <action B> named "B" must be executed _after_ <action A> named "A".
        <action B> named "B" must be executed _before_ <action C> named "C".
        <action A> named "A" must be executed _after_ <action C> named "C".

    """


def test_suite():
    # Override _ to not get translated test output
    import schooltool.app.main
    schooltool.app.main._ = lambda x: x

    optionflags = (doctest.ELLIPSIS |
                   doctest.NORMALIZE_WHITESPACE |
                   doctest.REPORT_NDIFF)
    return unittest.TestSuite([
                doctest.DocTestSuite(optionflags=doctest.ELLIPSIS),
                doctest.DocTestSuite('schooltool.app.main',
                                     optionflags=optionflags),
           ])

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Upgrade SchoolTool to generation 29.

Remove unused attributes from person schema.
"""
from zope.app.generations.utility import findObjectsProviding
from zope.app.publication.zopepublication import ZopePublication

from schooltool.app.interfaces import ISchoolToolApplication

def evolve(context):
    root = context.connection.root().get(ZopePublication.root_name, None)

    for app in findObjectsProviding(root, ISchoolToolApplication):
        persons = app['persons']
        for person in persons.values():
            del person.email
            del person.phone
            if hasattr(person, 'gradeclass'):
                del person.gradeclass



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for schooltool timetable schema views.
"""
import datetime
import unittest
import doctest
from pprint import pprint

from zope.component import provideAdapter, provideHandler, provideUtility
from zope.component import adapts
from zope.location.location import locate
from zope.interface import Interface, implements
from zope.interface import directlyProvides
from zope.interface import directlyProvidedBy
from zope.i18n import translate
from zope.publisher.browser import TestRequest
from zope.publisher.browser import BrowserView
from zope.traversing.interfaces import ITraversable
from zope.traversing.interfaces import IContainmentRoot
from zope.app.testing import setup
from zope.traversing import namespace
from zope.schema.vocabulary import getVocabularyRegistry

from schooltool.app.browser import testing
from schooltool.app.app import SimpleNameChooser
from schooltool.app.interfaces import IApplicationPreferences
from schooltool.testing import setup as sbsetup
from schooltool.testing.util import normalize_xml
from schooltool.testing.util import XMLCompareMixin
from schooltool.testing.util import NiceDiffsMixin
from schooltool.timetable.app import activityVocabularyFactory
from schooltool.timetable.daytemplates import DayTemplateSchedule
from schooltool.timetable.daytemplates import DayTemplate
from schooltool.timetable.daytemplates import TimeSlot
from schooltool.timetable.schedule import Period

try:
    from schooltool.timetable.schema import TimetableSchemaContainer
    from schooltool.timetable.model import WeeklyTimetableModel
    from schooltool.timetable.model import SequentialDaysTimetableModel
    from schooltool.timetable.interfaces import ITimetableModelFactory
    from schooltool.timetable.interfaces import ITimetableSchemaContainer
    from schooltool.timetable.browser.tests import test_timetable
    from schooltool.timetable.browser.tests.test_timetable import createDayTemplate
    from schooltool.timetable.browser.timetable import DayTemplatesTableSnippet
    from schooltool.timetable.browser.timetable import TimetableView
except:
    pass # XXX: tests not refactored yet


def setUp(test=None):
    testing.setUp(test)
    sbsetup.setUpApplicationPreferences()

    vr = getVocabularyRegistry()
    vr.register('schooltool.timetable.activityvocbulary',
                activityVocabularyFactory())

tearDown = testing.tearDown


def createSchema(*args):
    tts = test_timetable.createSchema(*args)
    tts.timezone = 'Asia/Tokyo'
    return tts


class TestAdvancedTimetableSchemaAdd(NiceDiffsMixin, unittest.TestCase):

    def setUp(self):
        setUp()
        self.app = sbsetup.setUpSchoolToolSite()
        self.ttschemas = TimetableSchemaContainer()
        locate(self.ttschemas, self.app, 'ttschemas')
        IApplicationPreferences(self.app).timezone = 'Asia/Tokyo'

        # Register the timetable models
        provideUtility(SequentialDaysTimetableModel,
                       ITimetableModelFactory,
                       'SequentialDaysTimetableModel')
        provideUtility(WeeklyTimetableModel,
                       ITimetableModelFactory,
                       'WeeklyTimetableModel')
        provideAdapter(SimpleNameChooser, (ITimetableSchemaContainer,))
        provideAdapter(lambda x: self.ttschemas, (Interface,), ITimetableSchemaContainer)

    def tearDown(self):
        tearDown()

    def createView(self, request=None):
        from schooltool.timetable.browser.timetable import \
             AdvancedTimetableSchemaAdd
        context = self.ttschemas
        context['default'] = createSchema(['Day 1'], ['Period 1'])
        if request is None:
            request = TestRequest()
        view = AdvancedTimetableSchemaAdd(context, request)
        return view

    def test(self):
        view = self.createView()
        result = view()
        self.assert_('name="field.title" size="20" type="text" value="default"'
                     in result)

    def test_with_data(self):
        request = TestRequest(form={'day1': 'Monday',
                                    'field.title': 'something',
                                    'model': 'SequentialDaysTimetableModel',
                                    'time1.period': 'Period 1',
                                    'time1.day0': '9:00-9:45',
                                    })
        view = self.createView(request)
        result = view()
        self.assertEquals(view.ttschema,
                          createSchema(['Monday'], ['Period 1']))
        self.assert_('value="something"' in result, result)
        self.assertEquals(view.title_widget.error(), '')
        self.assertEquals(view.model_name, 'SequentialDaysTimetableModel')
        self.assertEquals(view.model_error, None)
        self.assertEquals(view.day_templates,
                          {None: createDayTemplate([]),
                           0: createDayTemplate([(9, 0, 45)])})

    def test_creation(self):
        request = TestRequest(form={'day1': 'Monday',
                                    'field.title': 'some.thing',
                                    'model': 'SequentialDaysTimetableModel',
                                    'time1.period': 'Period 1',
                                    'time1.day0': '9:00-9:45',
                                    'CREATE': 'Create', })
        view = self.createView(request)
        result = view()
        self.assertEquals(view.model_error, None)
        self.assertEquals(request.response.getStatus(), 302)
        self.assertEquals(request.response.getHeader('location'),
                          'http://127.0.0.1/ttschemas')
        schema = view.context['something']
        self.assertEquals(schema, view.ttschema)
        self.assertEquals(schema.model.timetableDayIds, view.ttschema.keys())
        self.assertEquals(schema.model.dayTemplates, view.day_templates)
        self.assertEquals(schema.timezone, 'Asia/Tokyo')

    def test_model_error(self):
        request = TestRequest(form={'model': 'xxx',
                                    'CREATE': 'Create'})
        view = self.createView(request)
        view()
        self.assertEquals(view.model_error, "Please select a value")

    def test_model_error_ignored_unless_this_is_the_final_submit(self):
        view = self.createView(TestRequest(form={'field.title': 'Schema',
                                                 'model': 'xxx'}))
        view()
        self.assertEquals(view.model_error, None)

    def test_buildDayTemplates_empty(self):
        view = self.createView()
        dt = view._buildDayTemplates()
        self.assertEquals(dt, {None: createDayTemplate([])})

    def test_buildDayTemplates_simple(self):
        request = TestRequest(form={
            'time1.day0': '9:00',
            'time2.day0': '10:00-10:45',
            'time2.day6': '10:30-11:10',
            'field.duration': '45'})
        view = self.createView(request)
        dt = view._buildDayTemplates()
        self.assertEquals(dt,
                          {None: createDayTemplate([]),
                           0: createDayTemplate([(9, 0, 45),
                                                 (10, 0, 45)]),
                           6: createDayTemplate([(10, 30, 40)])})
        self.assert_(not view.discarded_some_periods)

    def test_buildDayTemplates_copy_day(self):
        request = TestRequest(form={'time1.period': 'Period 1',
                                     'time1.day0': '9:00-9:45',
                                     'time1.day1': '13:00-13:45',
                                     'time2.period': 'Period 2',
                                     'time2.day0': '10:00-10:45',
                                     'time2.day6': '10:30-11:10',
                                     'COPY_PERIODS_1': 'C'})
        view = self.createView(request)
        dt = view._buildDayTemplates()
        self.assertEquals(dt,
                          {None: createDayTemplate([]),
                           0: createDayTemplate([(9, 0, 45),
                                                 (10, 0, 45)]),
                           1: createDayTemplate([(9, 0, 45),
                                                 (10, 0, 45)]),
                           6: createDayTemplate([(10, 30, 40)])})

    def test_buildDayTemplates_copy_empty_day(self):
        request = TestRequest(form={
            'time1.period': 'Period 1',
            'time1.day0': '9:00-9:45',
            'time2.period': 'Period 2',
            'time2.day0': '10:00-10:45',
            'time2.day6': '10:30-11:10',
            'COPY_PERIODS_6': 'C'})
        view = self.createView(request)
        dt = view._buildDayTemplates()
        self.assertEquals(dt,
                          {None: createDayTemplate([]),
                           0: createDayTemplate([(9, 0, 45),
                                                 (10, 0, 45)])})

    def test_buildDayTemplates_copy_empty_day_over_empty_day(self):
        request = TestRequest(form={
            'time1.period': 'Period 1',
            'time1.day0': '9:00-9:45',
            'time2.period': 'Period 2',
            'time2.day0': '10:00-10:45',
            'time2.day6': '10:30-11:10',
            'COPY_PERIODS_4': 'C'})
        view = self.createView(request)
        dt = view._buildDayTemplates()
        self.assertEquals(dt,
                          {None: createDayTemplate([]),
                           0: createDayTemplate([(9, 0, 45),
                                                 (10, 0, 45)]),
                           6: createDayTemplate([(10, 30, 40)])})

    def test_buildDayTemplates_copy_first_day_ignored(self):
        request = TestRequest(form={
            'time1.period': 'Period 1',
            'time1.day0': '9:00-9:45',
            'time2.period': 'Period 2',
            'time2.day0': '10:00-10:45',
            'time2.day6': '10:30-11:10',
            'COPY_PERIODS_0': 'C'})
        view = self.createView(request)
        dt = view._buildDayTemplates()
        self.assertEquals(dt,
                          {None: createDayTemplate([]),
                           0: createDayTemplate([(9, 0, 45),
                                                 (10, 0, 45)]),
                           6: createDayTemplate([(10, 30, 40)])})

    def test_buildDayTemplates_errors(self):
        request = TestRequest(form={
            'time1.period': 'Period 1',
            'time1.day0': 'foo',
            # When duration_widget.value is None, both endpoints are required
            'time1.day1': '9:00'
            })
        view = self.createView(request)
        dt = view._buildDayTemplates()
        self.assertEquals(dt, {None: createDayTemplate([])})
        self.assert_(view.discarded_some_periods)

    def test_buildSchema_empty(self):
        view = self.createView()
        schema = view._buildSchema()
        self.assertEquals(schema, createSchema(['Day 1'], ['Period 1']))

    def test_buildSchema_from_request(self):
        request = TestRequest(form={
            'day1': 'Monday',
            'day2': ' Tuesday ',
            'day1.period1': 'A',
            'day1.period2': ' B ',
            'day2.period1': 'C',
            'day2.period2': ''})
        view = self.createView(request)
        schema = view._buildSchema()
        self.assertEquals(schema, createSchema(['Monday', 'Tuesday'],
                                               ['A', 'B'], ['C']))

    def test_buildSchema_empty_day(self):
        request = TestRequest(form={
            'day1': 'Monday',
            'day2': 'Tuesday',
            'day1.period1': 'A',
            'day1.period2': 'B',
            'day2.period1': '',
            'day2.period2': ''})
        view = self.createView(request)
        schema = view._buildSchema()
        self.assertEquals(schema, createSchema(['Monday', 'Tuesday'],
                                               ['A', 'B'], ['Period 1']))

    def test_buildSchema_repeated_day_name(self):
        request = TestRequest(form={
            'day1': 'D',
            'day2': 'D',
            'day3': 'D',
            'day1.period1': 'A',
            'day2.period1': 'B',
            'day3.period1': 'C'})
        view = self.createView(request)
        schema = view._buildSchema()
        self.assertEquals(schema, createSchema(['D', 'D (2)', 'D (3)'],
                                               ['A'], ['B'], ['C']))

    def test_buildSchema_repeated_period_nam(self):
        request = TestRequest(form={
            'day1': 'D',
            'day1.period1': 'A',
            'day1.period2': 'A'})
        view = self.createView(request)
        schema = view._buildSchema()
        self.assertEquals(schema, createSchema(['D'], ['A', 'A (2)']))

    def test_buildSchema_add_day(self):
        request = TestRequest(form={
            'day1': 'Monday',
            'ADD_DAY': "Add"})
        view = self.createView(request)
        schema = view._buildSchema()
        self.assertEquals(schema, createSchema(['Monday', 'Day 2'],
                                               ['Period 1'], ['Period 1']))

    def test_buildSchema_add_period(self):
        request = TestRequest(form={
            'day1': 'Monday',
            'day2': 'Tuesday',
            'day1.period1': 'A',
            'day1.period2': 'B',
            'day2.period1': 'C',
            'day2.period2': '',
            'ADD_PERIOD': "Add"})
        view = self.createView(request)
        schema = view._buildSchema()
        self.assertEquals(schema, createSchema(['Monday', 'Tuesday'],
                                               ['A', 'B', 'Period 3'], ['C']))

    def test_buildSchema_add_period_not_first_day(self):
        request = TestRequest(form={
            'day1': 'Monday',
            'day2': 'Tuesday',
            'day1.period1': 'A',
            'day2.period1': 'C',
            'day2.period2': 'D',
            'ADD_PERIOD': "Add"})
        view = self.createView(request)
        schema = view._buildSchema()
        self.assertEquals(schema, createSchema(['Monday', 'Tuesday'],
                                               ['A'], ['C', 'D', 'Period 3']))

    def test_buildSchema_delete_day(self):
        request = TestRequest(form={
            'day1': 'Day 1',
            'day2': 'Day 1',
            'day1.period1': 'A',
            'day1.period2': 'B',
            'day2.period1': 'C',
            'day2.period2': 'D',
            'DELETE_DAY_1': "Delete"})
        view = self.createView(request)
        schema = view._buildSchema()
        self.assertEquals(schema, createSchema(['Day 1'], ['C', 'D']))

    def test_buildSchema_copy_day(self):
        request = TestRequest(form={
            'day1': 'Day 1',
            'day2': 'Day 2',
            'day1.period1': 'A',
            'day1.period2': 'B',
            'day2.period1': 'C',
            'day2.period2': 'D',
            'COPY_DAY_1': "Copy"})
        view = self.createView(request)
        schema = view._buildSchema()
        self.assertEquals(schema, createSchema(['Day 1', 'Day 2'],
                                               ['A', 'B'], ['A', 'B']))

    def test_buildSchema_copy_first_day_ignored(self):
        request = TestRequest(form={
            'day1': 'Day 1',
            'day2': 'Day 2',
            'day1.period1': 'A',
            'day1.period2': 'B',
            'day2.period1': 'C',
            'day2.period2': 'D',
            'COPY_DAY_0': "Copy"})
        view = self.createView(request)
        schema = view._buildSchema()
        self.assertEquals(schema, createSchema(['Day 1', 'Day 2'],
                                               ['A', 'B'], ['C', 'D']))

    def test_all_periods(self):
        view = self.createView()
        view.ttschema = createSchema(['Day 1', 'Day 2', 'Day 3'],
                                     ['A', 'C'], ['B', 'D'], ['A', 'F'])
        self.assertEquals(view.all_periods(), ['A', 'C', 'B', 'D', 'F'])

    def test_slot_times(self):
        view = self.createView()
        view.ttschema = createSchema(['Day 1', 'Day 2', 'Day 3'],
                                     ['A', 'C'], ['B', 'D'], ['A', 'F', 'G'])
        view.day_templates = {}
        result = view.slot_times()
        # The length of the result is equal to the longest day in the tt.
        self.assertEquals(len(result), 3)
        for row in result:
            self.assertEquals(row, 7 * [None])

    def test_slot_times_with_data(self):
        view = self.createView()
        view.ttschema = createSchema(['Day 1', 'Day 2', 'Day 3'],
                                     ['A', 'C'], ['B', 'D'], ['A', 'F', 'G'])
        view.day_templates = {0: createDayTemplate([(9, 0, 45),
                                                    (10, 30, 40),
                                                    (11, 22, 33)]),
                              6: createDayTemplate([(8, 55, 45),
                                                    (0, 0, 24*60)])}
        times = view.slot_times()
        self.assertEquals(len(times), 3)
        self.assertEquals(times[0], ['09:00-09:45', None, None, None,
                                     None, None, '00:00-24:00'])
        self.assertEquals(times[1], ['10:30-11:10', None, None, None,
                                     None, None, '08:55-09:40'])
        self.assertEquals(times[2], ['11:22-11:55'] + [None] * 6)


def doctest_DayTemplatesTableSnippet():
    """Test for DayTemplatesTableSnippet.

    The purpose of this snippet is to iterate day templates and render
    them as a table.

        >>> schedule = DayTemplateSchedule()
        >>> schedule.initTemplates()

        >>> class Lesson(object):
        ...     def __init__(self, title, description):
        ...         self.title = title
        ...         self.description = description

        >>> templates = schedule.templates

    Let's add a few days.

        >>> def add_day(templates, title, lessons):
        ...     d = len(templates)
        ...     template = templates[u'%d' % d] = DayTemplate(title)
        ...     for n, lesson in enumerate(lessons):
        ...        template[u'%d' % n] = Lesson(u'lesson %d' % n, lesson)

        >>> lessons = ['math', 'english', 'history']
        >>> add_day(templates, u'Day 1', lessons)
        >>> add_day(templates, u'Day 2', lessons[1:])

    And create the view.

        >>> view = DayTemplatesTableSnippet(schedule, TestRequest())

        >>> [t.title for t in view.templates]
        [u'Day 1', u'Day 2']

    The view knows nothing about items in day templates, so we'll
    need to pass formatters that, given the item, return tuples
    of (title, value) for presentation.

        >>> formatter = lambda item: (item.__name__, item.title)

        >>> pprint(view.extractDays(formatter))
        [(u'Day 1', [(u'0', u'lesson 0'),
                     (u'1', u'lesson 1'),
                     (u'2', u'lesson 2')]),
         (u'Day 2', [(u'0', u'lesson 0'),
                     (u'1', u'lesson 1')])]

    Let's write a formatter better suited to represent our Lesson.

        >>> lesson_formatter = lambda t: (t.title, t.description)

        >>> days = view.extractDays(lesson_formatter)
        >>> pprint(days)
        [(u'Day 1',
          [(u'lesson 0', 'math'),
           (u'lesson 1', 'english'),
           (u'lesson 2', 'history')]),
         (u'Day 2',
          [(u'lesson 0', 'english'),
           (u'lesson 1', 'history')])]

    makeTable massages data into dict, better suited for page templates.

        >>> pprint(view.makeTable(days))
        {'col_width': '50%',
         'header': [u'Day 1', u'Day 2'],
         'rows': [({'title': u'lesson 0', 'value': 'math'},
                   {'title': u'lesson 0', 'value': 'english'}),
                  ({'title': u'lesson 1', 'value': 'english'},
                   {'title': u'lesson 1', 'value': 'history'}),
                  ({'title': u'lesson 2', 'value': 'history'},
                   {})],
         'td_width': '45%',
         'th_width': '5%'}

        >>> print normalize_xml(view(item_formatter=lesson_formatter))
        <table class="timetable">
          <tr>
            <th class="day" colspan="2" width="50%">
              Day 1
            </th>
            <th class="day" colspan="2" width="50%">
              Day 2
            </th>
          </tr>
          <tr>
            <th class="period" width="5%">
              lesson 0
            </th>
            <td class="activity" width="45%">
              math
            </td>
            <th class="period" width="5%">
              lesson 0
            </th>
            <td class="activity" width="45%">
              english
            </td>
          </tr>
          <tr>
            <th class="period" width="5%">
              lesson 1
            </th>
            <td class="activity" width="45%">
              english
            </td>
            <th class="period" width="5%">
              lesson 1
            </th>
            <td class="activity" width="45%">
              history
            </td>
          </tr>
          <tr>
            <th class="period" width="5%">
              lesson 2
            </th>
            <td class="activity" width="45%">
              history
            </td>
            <td class="activity" colspan="2" width="50%"/>
          </tr>
        </table>

    """


def doctest_TimetableView():
    """Test for TimetableView.

    This view queries snippets to render the period and time slot tables.

    The view is responsible for final formatting of period/time slot entries.
    To demonstrate this, we need stub day templates with samples of actual
    objects.

        >>> class IStubDayTemplates(Interface):
        ...     pass

        >>> class StubDayTemplates(object):
        ...     implements(IStubDayTemplates)
        ...     def __init__(self, sample=None):
        ...         self.sample = sample

        >>> period = Period(title='One', activity_type='lesson')
        >>> time_slot = TimeSlot(
        ...     datetime.time(12, 0),
        ...     datetime.timedelta(0, 2700),
        ...     activity_type='lunch')

        >>> class TimetableStub(object):
        ...     pass

        >>> timetable = TimetableStub()
        >>> timetable.periods = StubDayTemplates(sample=period)
        >>> timetable.time_slots = StubDayTemplates(sample=time_slot)

   Let's create our view.

        >>> view = TimetableView(timetable, TestRequest())

  The snippet is not registered yet, so not much action here.

        >>> view.periods_snippet()
        ''

        >>> view.time_snippet()
        ''

   Let's register a snippet that uses a fromatter from the view to render
   the sample object.

        >>> class TableSnippet(BrowserView):
        ...     adapts(IStubDayTemplates, TestRequest)
        ...     def __call__(self, item_formatter=None):
        ...         return item_formatter(self.context.sample)

        >>> from zope.publisher.interfaces.browser import IBrowserView

        >>> provideAdapter(TableSnippet,
        ...                provides=IBrowserView,
        ...                name="day_templates_table_snippet")

        >>> view.periods_snippet()
        ('One', u'Lesson')

        >>> view.time_snippet()
        ('12:00-12:45', u'Lunch')


    """


def doctest_SimpleTimetableSchemaAdd():
    r"""Doctest for the SimpleTimetableSchemaAdd view

        >>> provideUtility(WeeklyTimetableModel,
        ...                ITimetableModelFactory,
        ...                'WeeklyTimetableModel')
        >>> provideAdapter(SimpleNameChooser, (ITimetableSchemaContainer,))
        >>> schemas = TimetableSchemaContainer()
        >>> provideAdapter(lambda x: schemas,
        ...                (Interface,), ITimetableSchemaContainer)

    Suppose we have a SchoolTool instance, and create a view for its
    timetable schemas container:

        >>> app = sbsetup.setUpSchoolToolSite()
        >>> IApplicationPreferences(app).timezone = 'Asia/Tokyo'

        >>> request = TestRequest()
        >>> from schooltool.timetable.browser.timetable import \
        ...      SimpleTimetableSchemaAdd
        >>> view = SimpleTimetableSchemaAdd(schemas, request)

        >>> locate(schemas, app, 'ttschemas')

    Let's render it.  There is a widget for title there:

        >>> print view()
        <BLANKLINE>
        ...
                  <div class="field"><input class="textType"
                             id="field.title"
                             name="field.title" size="20" type="text"
                             value="default"  /></div>
        ...
          <tr>
            <th>
              <div><input class="textType" id="field.period_name_1"
                          name="field.period_name_1" size="20" type="text"
                          value=""  /></div>
            </th>
            <th>
              <div><input class="textType" id="field.period_start_1"
                          name="field.period_start_1" size="20"
                          type="text" value="" /></div>
            </th>
            <th>
              <div><input class="textType" id="field.period_finish_1"
                          name="field.period_finish_1" size="20"
                          type="text" value="" /></div>
            </th>
          </tr>
        ...

    getPeriods returns None, as the form is not yet filled:

        >>> view.getPeriods()
        []

    Now, let's create a simple case with all the fields filled:

        >>> request = TestRequest(form={'field.title': 'default',
        ...                             'field.period_name_1': 'Period 1',
        ...                             'field.period_start_1': '9:00',
        ...                             'field.period_finish_1': '9:45',
        ...                             'field.period_name_2': 'Period 2',
        ...                             'field.period_start_2': '10:00',
        ...                             'field.period_finish_2': '10:45',
        ...                             'field.period_name_3': '',
        ...                             'field.period_name_4': '',
        ...                             'CREATE': 'Go'
        ...                            })
        >>> view = SimpleTimetableSchemaAdd(schemas, request)

    getPeriods should extract a list of periods:

        >>> pprint(view.getPeriods())
        [(u'Period 1',
          datetime.time(9, 0),
          datetime.timedelta(0, 2700)),
         (u'Period 2',
          datetime.time(10, 0),
          datetime.timedelta(0, 2700))]

    If we call the view, a new timetable schema is created:

        >>> result = view()
        >>> list(view.context.keys())
        [u'default']
        >>> schema = view.context['default']
        >>> print " ".join(schema.day_ids)
        Monday Tuesday Wednesday Thursday Friday
        >>> print ", ".join(schema['Monday'].periods)
        Period 1, Period 2

    All days are the same

        >>> for day_id, day in schema.items():
        ...     assert day == schema['Monday']

    The schema uses the weekly timetable model

        >>> print schema.model.__class__.__name__
        WeeklyTimetableModel
        >>> print " ".join(schema.model.timetableDayIds)
        Monday Tuesday Wednesday Thursday Friday
        >>> for period in schema.model.dayTemplates[None]:
        ...     print period.tstart, period.duration
        09:00:00 0:45:00
        10:00:00 0:45:00
        >>> schema.timezone
        'Asia/Tokyo'

    We should get redirected to the ttschemas index:

        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('location')
        'http://127.0.0.1/ttschemas'

    If period names are not provided, use start times:

        >>> request = TestRequest(form={'field.title': 'default',
        ...                             'field.period_name_1': '',
        ...                             'field.period_start_1': '9:00',
        ...                             'field.period_finish_1': '9:45',
        ...                             'field.period_name_2': '',
        ...                             'field.period_start_2': '10:00',
        ...                             'field.period_finish_2': '10:45',
        ...                             'CREATE': 'Go'
        ...                            })
        >>> view = SimpleTimetableSchemaAdd(schemas, request)

        >>> pprint(view.getPeriods())
        [(u'9:00',
          datetime.time(9, 0),
          datetime.timedelta(0, 2700)),
         (u'10:00',
          datetime.time(10, 0),
          datetime.timedelta(0, 2700))]

    If a cancel button is pressed, nothing is done and the user is
    redirected to ttschemas index:

        >>> request = TestRequest(form={'field.title': 'default2',
        ...                             'field.period_name_1': 'Period 1',
        ...                             'field.period_start_1': '9:00',
        ...                             'field.period_finish_1': '9:45',
        ...                             'field.period_name_2': 'Period 2',
        ...                             'field.period_start_2': '10:00',
        ...                             'field.period_finish_2': '10:45',
        ...                             'field.period_name_3': '',
        ...                             'field.period_name_4': '',
        ...                             'CANCEL': 'Cancel'
        ...                            })
        >>> view = SimpleTimetableSchemaAdd(schemas, request)
        >>> result = view()
        >>> list(view.context.keys())
        [u'default']

        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('location')
        'http://127.0.0.1/ttschemas'

    If there's a period skipped in a form, consequent periods are not included:

        >>> request = TestRequest(form={'field.title': 'default',
        ...                             'field.period_name_1': 'Period 1',
        ...                             'field.period_start_1': '9:00',
        ...                             'field.period_finish_1': '9:45',
        ...                             'field.period_name_3': 'Period 2',
        ...                             'field.period_start_3': '10:00',
        ...                             'field.period_finish_3': '10:45',
        ...                             'field.period_name_4': '',
        ...                            })
        >>> view = SimpleTimetableSchemaAdd(schemas, request)
        >>> view.getPeriods()
        [(u'Period 1',
          datetime.time(9, 0),
          datetime.timedelta(0, 2700)),
         (u'Period 2',
          datetime.time(10, 0),
          datetime.timedelta(0, 2700))]

    If a period does not have a start time or end time specified, it
    is skipped:

        >>> request = TestRequest(form={'field.title': 'default',
        ...                             'field.period_name_1': 'Period 1',
        ...                             'field.period_start_1': '9:00',
        ...                             'field.period_finish_1': '9:45',
        ...                             'field.period_name_2': 'Period 2',
        ...                             'field.period_start_2': '',
        ...                             'field.period_finish_2': '10:45',
        ...                             'field.period_name_3': 'Period 3',
        ...                             'field.period_start_3': '11:00',
        ...                             'field.period_finish_3': '',
        ...                            })
        >>> view = SimpleTimetableSchemaAdd(schemas, request)
        >>> view.getPeriods()
        [(u'Period 1',
          datetime.time(9, 0),
          datetime.timedelta(0, 2700))]

    Incorrect start and end times are handled gracefully:

        >>> request = TestRequest(form={'field.title': 'default',
        ...                             'field.period_name_1': 'Period 1',
        ...                             'field.period_start_1': '9:00',
        ...                             'field.period_finish_1': '9:45',
        ...                             'field.period_name_2': '',
        ...                             'field.period_start_2': '10h',
        ...                             'field.period_finish_2': '',
        ...                            })
        >>> view = SimpleTimetableSchemaAdd(schemas, request)

    getPeriods fails:

        >>> view.getPeriods()
        [(u'Period 1', datetime.time(9, 0), datetime.timedelta(0, 2700))]

    The widgets responsible get an error set on them:

        >>> print view()
        <BLANKLINE>
        ...
            <div class="error">Please use HH:MM format for period start
                               and end times</div>
        ...

        >>> request.response.getStatus() != 302
        True


   One can provide the same title more than once (thought it is not
   advised to do so):

        >>> request = TestRequest(form={
        ...                             'field.title': 'already',
        ...                             'field.period_name_1': 'p1',
        ...                             'field.period_start_1': '9:00',
        ...                             'field.period_finish_1': '10:00',
        ...                             'CREATE': 'Create'
        ...                            })
        >>> view = SimpleTimetableSchemaAdd(schemas, request)
        >>> from schooltool.timetable.schema import TimetableSchema
        >>> schemas['already'] = TimetableSchema([])
        >>> result = view()
        >>> request.response.getStatus() == 302
        True
        >>> 'already-2' in schemas
        True

    """


def doctest_SimpleTimetableSchemaAdd_errors():
    r"""Doctest for the SimpleTimetableSchemaAdd view

        >>> from schooltool.timetable.model import WeeklyTimetableModel
        >>> from schooltool.timetable.interfaces import ITimetableModelFactory
        >>> provideUtility(WeeklyTimetableModel,
        ...                ITimetableModelFactory,
        ...                'WeeklyTimetableModel')

    Suppose we have a SchoolTool instance, and create a view for its
    timetable schemas container:

        >>> from schooltool.timetable.schema import TimetableSchemaContainer
        >>> app = sbsetup.setUpSchoolToolSite()
        >>> schemas = TimetableSchemaContainer()

    No name specified:

        >>> request = TestRequest(form={
        ...                             'field.title': '',
        ...                             'field.period_name_1': 'p1',
        ...                             'field.period_start_1': '9:00',
        ...                             'field.period_finish_1': '10:00',
        ...                             'CREATE': 'Create'
        ...                            })
        >>> from schooltool.timetable.browser.timetable import \
        ...      SimpleTimetableSchemaAdd
        >>> view = SimpleTimetableSchemaAdd(schemas, request)
        >>> print view()
        <BLANKLINE>
        ...
                 <div class="label">
                    <label for="field.title" title="">Title</label>
                  </div>
                  <span class="error">Required input is missing.</span>
                  <div class="field"><input class="textType" id="field.title"
                                            name="field.title" size="20"
                                            type="text" value=""  /></div>
        ...
          <tr>
            <th>
              <div><input class="textType" id="field.period_name_1"
                          name="field.period_name_1" size="20" type="text"
                          value="p1"  /></div>
            </th>
            <th>
              <div><input class="textType" id="field.period_start_1"
                          name="field.period_start_1" size="20"
                          type="text" value="9:00" /></div>
            </th>
            <th>
              <div><input class="textType" id="field.period_finish_1"
                          name="field.period_finish_1" size="20"
                          type="text" value="10:00" /></div>
            </th>
          </tr>
        ...

        >>> request.response.getStatus() != 302
        True

    No periods:

        >>> request = TestRequest(form={
        ...                             'field.title': 'empty',
        ...                             'field.period_name_1': '',
        ...                             'field.period_start_1': '',
        ...                             'field.period_finish_1': '',
        ...                             'CREATE': 'Create'
        ...                            })
        >>> view = SimpleTimetableSchemaAdd(schemas, request)
        >>> print view()
        <BLANKLINE>
        ...
            <div class="error">You must specify at least one period.</div>
        ...
                  <div class="label">
                    <label for="field.title" title="">Title</label>
                  </div>
                  <div class="field"><input class="textType" id="field.title"
                                            name="field.title" size="20"
                                            type="text" value="empty"
                                            /></div>
        ...

        >>> request.response.getStatus() != 302
        True

    """


def doctest_TimetableSchemaContainerView():
    r"""A test for TimetableSchemaContainer view

    We will need an application:

        >>> app = sbsetup.createSchoolToolApplication()
        >>> schemas = TimetableSchemaContainer()
        >>> provideAdapter(lambda x: schemas,
        ...                adapts=[Interface],
        ...                provides=ITimetableSchemaContainer)

    Some timetable schemas:

        >>> from schooltool.timetable.schema import TimetableSchema
        >>> schemas["schema1"] = TimetableSchema([])
        >>> schemas["schema2"] = TimetableSchema([])

    Let's create our view:

        >>> from schooltool.timetable.browser.timetable import \
        ...      TimetableSchemaContainerView
        >>> from zope.publisher.browser import TestRequest
        >>> view = TimetableSchemaContainerView(schemas,
        ...                                     TestRequest())

    The default ttschema id should be "schema1":

        >>> schemas.default_id
        'schema1'

    If the view is submited without any data - the default ttschema
    should not change:

        >>> view.update()
        ''
        >>> schemas.default_id
        'schema1'

    We can change the default schema:

        >>> view.request = TestRequest(form={
        ...                                  'ttschema': 'schema2',
        ...                                  'UPDATE_SUBMIT': 'Change'
        ...                                 })
        >>> view.update()
        ''
        >>> schemas.default_id
        'schema2'

    We can set the default_id to none:

        >>> view.request = TestRequest(form={
        ...                                  'ttschema': '',
        ...                                  'UPDATE_SUBMIT': 'Change'
        ...                                 })
        >>> view.update()
        ''
        >>> schemas.default_id is None
        True

    """


def doctest_TimetableDependentDeleteView():
    r"""Tests for TimetableDependentDeleteView.

    This view is used to delete school timetables and terms, and
    recursively deletes all timetables that use the object to be
    deleted.

    First, let's set up:

        >>> from schooltool.timetable import TimetablesAdapter
        >>> from schooltool.timetable.interfaces import ITimetables
        >>> setup.placefulSetUp()
        >>> provideAdapter(TimetablesAdapter)
        >>> setup.setUpAnnotations()
        >>> from schooltool.timetable.schema import clearTimetablesOnDeletion
        >>> from schooltool.timetable.interfaces import ITimetableSchema
        >>> from zope.lifecycleevent.interfaces import IObjectRemovedEvent
        >>> provideHandler(clearTimetablesOnDeletion,
        ...                (ITimetableSchema, IObjectRemovedEvent))

    Now, let's create a couple of timetables to operate on:

        >>> from schooltool.timetable.schema import TimetableSchema
        >>> from schooltool.timetable.schema import TimetableSchemaDay
        >>> from schooltool.timetable.interfaces import IOwnTimetables
        >>> app = sbsetup.setUpSchoolToolSite()
        >>> schemas = TimetableSchemaContainer()
        >>> locate(schemas, app, "ttschemas")
        >>> directlyProvides(app, directlyProvidedBy(app) + IOwnTimetables)

        >>> days = ('A', 'B')
        >>> periods1 = ('Green', 'Blue')
        >>> tts = TimetableSchema(days)
        >>> tts["A"] = TimetableSchemaDay(periods1)
        >>> tts["B"] = TimetableSchemaDay(periods1)

        >>> days = ('C', 'D')
        >>> tts2 = TimetableSchema(days)
        >>> tts2["C"] = TimetableSchemaDay(periods1)
        >>> tts2["D"] = TimetableSchemaDay(periods1)

        >>> schemas['simple'] = tts
        >>> schemas['other'] = tts2

    Let's create a couple of timetables on the application:

        >>> ITimetables(app).timetables['2006.simple'] = tts.createTimetable(None)
        >>> ITimetables(app).timetables['2006.other'] = tts2.createTimetable(None)

    Now, we can run the view:

        >>> from schooltool.timetable.browser.timetable \
        ...     import TimetableDependentDeleteView
        >>> request = TestRequest(form={'delete.simple': 'on',
        ...                             'CONFIRM': 'Confirm'})
        >>> view = TimetableDependentDeleteView(schemas, request)
        >>> view.update()

    The timetable schema is deleted:

        >>> list(schemas.keys())
        [u'other']

    Also, the dependent timetables have been deleted:

        >>> ITimetables(app).timetables.keys()
        ['2006.other']

    The user is redirected to the school timetable index view:

        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('Location')
        'http://127.0.0.1/ttschemas'

    The user can also cancel the deletion:

        >>> from schooltool.timetable.browser.timetable \
        ...     import TimetableDependentDeleteView
        >>> request = TestRequest(form={'delete.other': 'on',
        ...                             'CANCEL': 'Cancel'})
        >>> view = TimetableDependentDeleteView(schemas, request)
        >>> view.update()

    In that case, the data is unchanged:

        >>> list(schemas.keys())
        [u'other']
        >>> ITimetables(app).timetables.keys()
        ['2006.other']

    The user is also redirected to the school timetable index view:

        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('Location')
        'http://127.0.0.1/ttschemas'
    """


class TimetableSchemaMixin(object):

    schema_xml = """
        <timetable xmlns="http://schooltool.org/ns/timetable/0.1">
          <title>Title</title>
          <timezone name="Europe/Vilnius"/>
          <model factory="SequentialDaysTimetableModel">
            <daytemplate>
              <used when="default" />
              <period id="A" tstart="9:00" duration="60" />
              <period id="C" tstart="9:00" duration="60" />
              <period id="B" tstart="10:00" duration="60" />
              <period id="D" tstart="10:00" duration="60" />
            </daytemplate>
            <daytemplate>
              <used when="Friday Thursday" />
              <period id="A" tstart="8:00" duration="60" />
              <period id="C" tstart="8:00" duration="60" />
              <period id="B" tstart="11:00" duration="60" />
              <period id="D" tstart="11:00" duration="60" />
            </daytemplate>
            <daytemplate>
              <used when="2005-07-07" />
              <period id="A" tstart="8:00" duration="30" />
              <period id="B" tstart="8:30" duration="30" />
              <period id="C" tstart="9:00" duration="30" />
              <period id="D" tstart="9:30" duration="30" />
            </daytemplate>
            <day when="2005-07-08" id="Day 2" />
            <day when="2005-07-09" id="Day 1" />
          </model>
          <day id="Day 1">
            <period id="A" homeroom="">
            </period>
            <period id="B">
            </period>
          </day>
          <day id="Day 2">
            <period id="C">
            </period>
            <period id="D">
            </period>
          </day>
        </timetable>
        """

    schema_without_title_xml = schema_xml.replace("<title>Title</title>", "")

    def setUp(self):
        from schooltool.timetable.model import SequentialDaysTimetableModel
        from schooltool.timetable.model import SequentialDayIdBasedTimetableModel
        from schooltool.timetable.interfaces import ITimetableModelFactory

        self.app = sbsetup.createSchoolToolApplication()
        self.schemaContainer = TimetableSchemaContainer()

        setup.placelessSetUp()
        setup.setUpTraversal()

        provideUtility(SequentialDaysTimetableModel,
                       ITimetableModelFactory,
                       "SequentialDaysTimetableModel")

        provideUtility(SequentialDayIdBasedTimetableModel,
                       ITimetableModelFactory,
                       "SequentialDayIdBasedTimetableModel")

        provideAdapter(namespace.view, (None,), ITraversable, 'view')

        directlyProvides(self.schemaContainer, IContainmentRoot)

    def tearDown(self):
        setup.placelessTearDown()

    def createEmptySchema(self):
        from schooltool.timetable.schema import TimetableSchemaDay
        from schooltool.timetable.schema import TimetableSchema
        schema = TimetableSchema(['Day 1', 'Day 2'])
        schema['Day 1'] = TimetableSchemaDay(['A', 'B'], ['A'])
        schema['Day 2'] = TimetableSchemaDay(['C', 'D'])
        schema.title = "A Schema"
        return schema

    def createExtendedSchema(self):
        from schooltool.timetable.model import SequentialDaysTimetableModel
        from schooltool.timetable import SchooldaySlot, SchooldayTemplate
        from datetime import time, timedelta, date

        tt = self.createEmptySchema()

        tt.timezone = 'Europe/Vilnius'

        hour = timedelta(minutes=60)
        half = timedelta(minutes=30)

        day_template1 = SchooldayTemplate()
        day_template1.add(SchooldaySlot(time(9, 0), hour))
        day_template1.add(SchooldaySlot(time(10, 0), hour))
        day_template1.add(SchooldaySlot(time(9, 0), hour))
        day_template1.add(SchooldaySlot(time(10, 0), hour))

        day_template2 = SchooldayTemplate()
        day_template2.add(SchooldaySlot(time(8, 0), hour))
        day_template2.add(SchooldaySlot(time(11, 0), hour))
        day_template2.add(SchooldaySlot(time(8, 0), hour))
        day_template2.add(SchooldaySlot(time(11, 0), hour))

        tm = SequentialDaysTimetableModel(['Day 1', 'Day 2'],
                                          {None: day_template1,
                                           3: day_template2,
                                           4: day_template2})
        tt.model = tm

        short_template = [
            ('A', SchooldaySlot(time(8, 0), half)),
            ('B', SchooldaySlot(time(8, 30), half)),
            ('C', SchooldaySlot(time(9, 0), half)),
            ('D', SchooldaySlot(time(9, 30), half))]
        tt.model.exceptionDays[date(2005, 7, 7)] = short_template
        tt.model.exceptionDayIds[date(2005, 7, 8)] = 'Day 2'
        tt.model.exceptionDayIds[date(2005, 7, 9)] = 'Day 1'
        return tt


class TestTimetableSchemaXMLView(TimetableSchemaMixin, XMLCompareMixin,
                              unittest.TestCase):

    empty_xml = """
        <timetable xmlns="http://schooltool.org/ns/timetable/0.1">
          <title>A Schema</title>
          <timezone name="Europe/Vilnius"/>
          <model factory="SequentialDaysTimetableModel">
            <daytemplate>
              <used when="2005-07-07"/>
              <period duration="30" id="A" tstart="08:00"/>
              <period duration="30" id="B" tstart="08:30"/>
              <period duration="30" id="C" tstart="09:00"/>
              <period duration="30" id="D" tstart="09:30"/>
            </daytemplate>
            <daytemplate>
              <used when="Friday Thursday"/>
              <period duration="60" tstart="08:00"/>
              <period duration="60" tstart="11:00"/>
            </daytemplate>
            <daytemplate>
              <used when="default"/>
              <period duration="60" tstart="09:00"/>
              <period duration="60" tstart="10:00"/>
            </daytemplate>
            <day when="2005-07-08" id="Day 2" />
            <day when="2005-07-09" id="Day 1" />
          </model>
          <day id="Day 1">
            <period id="A" homeroom="">
            </period>
            <period id="B">
            </period>
          </day>
          <day id="Day 2">
            <period id="C">
            </period>
            <period id="D">
            </period>
          </day>
        </timetable>
        """

    def test_get(self):
        from schooltool.timetable.browser.timetable import TimetableSchemaXMLView
        request = TestRequest()
        view = TimetableSchemaXMLView(self.createExtendedSchema(), request)

        result = view()
        self.assertEquals(request.response.getHeader('content-type'),
                          "text/xml; charset=UTF-8")
        self.assertEqualsXML(result, self.empty_xml)


class DayIdBasedModelMixin:

    empty_xml = """
        <timetable xmlns="http://schooltool.org/ns/timetable/0.1">
          <title>Title</title>
          <timezone name="UTC"/>
          <model factory="SequentialDayIdBasedTimetableModel">
            <daytemplate>
              <used when="Day 1"/>
              <period duration="60" tstart="08:00"/>
              <period duration="60" tstart="11:00"/>
            </daytemplate>
            <daytemplate>
              <used when="Day 2"/>
              <period duration="60" tstart="09:00"/>
              <period duration="60" tstart="10:00"/>
            </daytemplate>
            <day when="2005-07-08" id="Day 2" />
            <day when="2005-07-09" id="Day 1" />
          </model>
          <day id="Day 1">
            <period id="A" homeroom="">
            </period>
            <period id="B">
            </period>
          </day>
          <day id="Day 2">
            <period id="C">
            </period>
            <period id="D">
            </period>
          </day>
        </timetable>
        """

    def createExtendedSchema(self):
        from schooltool.timetable.schema import TimetableSchemaDay
        from schooltool.timetable.schema import TimetableSchema
        from schooltool.timetable.model import SequentialDayIdBasedTimetableModel
        from schooltool.timetable import SchooldaySlot, SchooldayTemplate
        from datetime import time, timedelta, date

        tt = TimetableSchema(['Day 1', 'Day 2'])
        tt['Day 1'] = TimetableSchemaDay(['A', 'B'], ['A'])
        tt['Day 2'] = TimetableSchemaDay(['C', 'D'])
        tt.title = "Title"

        hour = timedelta(minutes=60)
        half = timedelta(minutes=30)

        day_template1 = SchooldayTemplate()
        day_template1.add(SchooldaySlot(time(8, 0), hour))
        day_template1.add(SchooldaySlot(time(11, 0), hour))
        day_template1.add(SchooldaySlot(time(8, 0), hour))
        day_template1.add(SchooldaySlot(time(11, 0), hour))

        day_template2 = SchooldayTemplate()
        day_template2.add(SchooldaySlot(time(9, 0), hour))
        day_template2.add(SchooldaySlot(time(10, 0), hour))
        day_template2.add(SchooldaySlot(time(9, 0), hour))
        day_template2.add(SchooldaySlot(time(10, 0), hour))

        tm = SequentialDayIdBasedTimetableModel(['Day 1', 'Day 2'],
                                                {'Day 1': day_template1,
                                                 'Day 2': day_template2})
        tt.model = tm

        tt.model.exceptionDayIds[date(2005, 7, 8)] = 'Day 2'
        tt.model.exceptionDayIds[date(2005, 7, 9)] = 'Day 1'
        return tt


class TestTimetableSchemaXMLViewDayIdBased(DayIdBasedModelMixin,
                                           TestTimetableSchemaXMLView):
    pass


def test_suite():
    suite = unittest.TestSuite()
    optionflags = (doctest.ELLIPSIS | doctest.REPORT_NDIFF |
                   doctest.NORMALIZE_WHITESPACE)
    # XXX: tests not refactored yet
    #suite.addTest(doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
    #                                   optionflags=optionflags))
    #suite.addTest(unittest.makeSuite(TestAdvancedTimetableSchemaAdd))
    #suite.addTest(unittest.makeSuite(TestTimetableSchemaXMLView))
    #suite.addTest(unittest.makeSuite(TestTimetableSchemaXMLViewDayIdBased))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Functional Testing Utilities
"""

import os
import unittest
import doctest

from zope.testbrowser.testing import Browser
from zope.testing.server import startServer
from zope.app.testing.functional import HTTPCaller
from zope.app.testing.functional import ZCMLLayer as _ZCMLLayer
from zope.app.testing.functional import FunctionalDocFileSuite
from zope.app.appsetup.interfaces import IDatabaseOpenedEvent
import zope.event

from schooltool.testing.analyze import queryHTML
from schooltool.testing import analyze


def find_ftesting_zcml():
    """Find ftesting.zcml to be used for SchoolTool functional tests."""
    dir = os.path.abspath(os.path.dirname(__file__))
    while True:
        filename = os.path.join(dir, 'etc', 'ftesting.zcml')
        if os.path.exists(filename):
            return filename # We are testing in an instance
        dir = os.path.dirname(dir)
        if dir == os.path.dirname(dir): # we're looping at the filesystem root
            raise RuntimeError("I can't find ftesting.zcml!")


class ZCMLLayer(_ZCMLLayer):

    def __init__(self, *args, **kwargs):
        kwargs['allow_teardown'] = True
        _ZCMLLayer.__init__(self, *args, **kwargs)

    def setUp(self):
        # SchoolTool needs to bootstrap the database first, before the Zope 3
        # IDatabaseOpenedEvent gets a chance to create its own root folder and
        # stuff.  Unfortunatelly, we cannot install a IDatabaseOpenedEvent
        # subscriber via ftesting.zcml and ensure it will get called first.
        # Instead we place our own subscriber directly into zope.event.subscribers,
        # where it gets a chance to intercept IDatabaseOpenedEvent before the
        # Zope 3 event dispatcher sees it.

        def install_db_bootstrap_hook():
            """Install schooltool_db_setup into zope.event.subscribers."""
            zope.event.subscribers.insert(0, schooltool_db_setup)

        def uninstall_db_bootstrap_hook():
            """Remove schooltool_db_setup from zope.event.subscribers."""
            zope.event.subscribers.remove(schooltool_db_setup)

        def schooltool_db_setup(event):
            """IDatabaseOpenedEvent handler that bootstraps SchoolTool."""
            if IDatabaseOpenedEvent.providedBy(event):
                import schooltool.app.main
                server = schooltool.app.main.SchoolToolServer()
                server.initializePreferences = lambda app: None
                server.bootstrapSchoolTool(event.database)
                server.startApplication(event.database)

        install_db_bootstrap_hook()
        try:
            _ZCMLLayer.setUp(self)
        finally:
            uninstall_db_bootstrap_hook()


class TestBrowser(Browser):

    username = None
    password = None

    def __init__(self, username=None, password=None, url='http://localhost/'):
        super(TestBrowser, self).__init__()
        self.username = username
        self.password = password
        if username and password:
            self.addHeader('Authorization',
                           'Basic %s:%s' % (self.username, self.password))
        self.handleErrors = False
        self.open(url)

    def serve(self, url=None, port=8000):
        if url is None:
            url = self.url
        startServer(HTTPCaller(), url, self.username, self.password, port=port)

    def queryHTML(self, query):
        return queryHTML(query, self.contents)

    def printQuery(self, query, skip_inner_blank=False):
        for item in queryHTML(query, self.contents):
            if item.strip():
                if skip_inner_blank:
                    result = str(item.strip()).splitlines()
                    for line in result:
                        line = line.strip()
                        if line:
                            print line
                else:
                    print item.strip()


def collect_txt_ftests(package=None, level=None, layer=None, filenames=None,
                       suite_factory=None):
    """Collect all functional doctest files in a given package.

    If `package` is None, looks up the call stack for the right module.

    Returns a unittest.TestSuite.
    """
    testdir = os.path.dirname(package.__file__)
    if filenames is None:
        filenames = [fn for fn in os.listdir(testdir)
                     if fn.endswith('.txt') and not fn.startswith('.')]
    suites = []
    for filename in filenames:
        suite = suite_factory(
            filename, package=package)
        if level is not None:
            suite.level = level
        if layer is None:
            raise ValueError("ftests must specify an ftesting.zcml.")
        suite.layer = layer
        suites.append(suite)
    return unittest.TestSuite(suites)


def collect_ftests(package=None, level=None, layer=None, filenames=None,
                   extra_globs=None):
    package = doctest._normalize_module(package)
    extra_globs = extra_globs or {}
    extra_globs.update({'analyze': analyze,
                        'Browser': TestBrowser})
    def make_suite(filename, package=None):
        optionflags = (doctest.ELLIPSIS | doctest.REPORT_NDIFF |
                       doctest.NORMALIZE_WHITESPACE |
                       doctest.REPORT_ONLY_FIRST_FAILURE)
        suite = FunctionalDocFileSuite(filename, package=package,
                                       optionflags=optionflags,
                                       globs=extra_globs)
        return suite
    return collect_txt_ftests(package=package, level=level,
                              layer=layer, filenames=filenames,
                              suite_factory=make_suite)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Schooltool content providers.
"""
from zope.component import adapts, adapter, queryMultiAdapter, getAdapter
from zope.interface import implements, implementer
import zope.contentprovider.interfaces
from zope.contentprovider.interfaces import ContentProviderLookupError
from zope.contentprovider.tales import addTALNamespaceData
from zope.contentprovider.provider import ContentProviderBase
from zope.event import notify
from zope.location import ILocation
from zope.proxy.decorator import SpecificationDecoratorBase
from zope.publisher.interfaces import NotFound
from zope.publisher.browser import BrowserPage
from zope.security.proxy import removeSecurityProxy
from zope.tales.interfaces import ITALESFunctionNamespace

from schooltool.skin.flourish import interfaces
from schooltool.securitypolicy.interfaces import ICrowd


class ContentProvider(ContentProviderBase, BrowserPage):
    """Base SchoolTool content provider class."""

    implements(interfaces.IContentProvider)

    @property
    def view(self):
        return self.__parent__

    @view.setter
    def view(self, value):
        self.__parent__ = value

    def __call__(self, *args, **kw):
        event = zope.contentprovider.interfaces.BeforeUpdateEvent
        notify(event(self, self.request))
        self.update()
        return self.render(*args, **kw)


class ContentProviderProxy(SpecificationDecoratorBase):
    """A content provider proxy that mimics behaviour of
    zope.contentprovider.tales.TALESProviderExpression
    """
    adapts(zope.contentprovider.interfaces.IContentProvider)
    implements(interfaces.IContentProvider)

    @property
    def view(self):
        return self.__parent__

    def browserDefault(self, request):
        return self, ()

    def publishTraverse(self, request, name):
        raise NotFound(self, name, request)

    def __call__(self, *args, **kw):
        event = zope.contentprovider.interfaces.BeforeUpdateEvent
        notify(event(self, self.request))
        self.update()
        return self.render(*args, **kw)


class ContentProviders(object):
    implements(interfaces.IContentProviders)

    def __init__(self, context, request, view):
        self.context = context
        self.request = request
        self.view = view
        self.cache = {}

    def __contains__(self, name):
        return self.get(name) is not None

    def get(self, name, default=None):
        if name not in self.cache:
            self[name] = self.lookup(name)
        ob = self.cache.get(name)
        if ob is None:
            return default
        return ob

    def __getitem__(self, name):
        ob = self.get(name, None)
        if ob is None:
            raise KeyError(name)
        return ob

    def __setitem__(self, name, provider):
        self.cache[name] = provider

    def lookup(self, name):
        provider = queryMultiAdapter(
            (self.context, self.request, self.view),
            interfaces.IContentProvider, name)
        if provider is None:
            provider = queryMultiAdapter(
                (self.context, self.request, self.view),
                zope.contentprovider.interfaces.IContentProvider, name)
            if provider is not None:
                provider = interfaces.IContentProvider(provider, None)
        if (provider is not None and
            ILocation.providedBy(provider)):
            unproxied = removeSecurityProxy(provider)
            unproxied.__name__ = name
        return provider

    def get_providers(self, direct=True, adapted=True):
        """Get all content providers.
        Use this for debugging purposes only.
        """
        if direct:
            direct_providers = list(zope.component.getAdapters(
                    (self.context, self.request, self.view),
                    interfaces.IContentProvider))
        else:
            direct_providers = []
        if adapted:
            indirect = list(zope.component.getAdapters(
                    (self.context, self.request, self.view),
                    zope.contentprovider.interfaces.IContentProvider))
            adapted_providers = [
                (name, interfaces.IContentProvider(provider, None))
                for name, provider in indirect]
        else:
            adapted_providers = []
        result = dict(adapted_providers)
        result.update(dict(direct_providers))
        return result

    def traverse(self, name, furtherPath):
        provider = self.get(name, None)
        if provider is None:
            raise ContentProviderLookupError(self.context, name)
        return provider


class TALESAwareContentProviders(ContentProviders):
    implements(ITALESFunctionNamespace)

    engine = None

    def setEngine(self, engine):
        self.engine = engine

    def traverse(self, name, furtherPath):
        provider = ContentProviders.traverse(self, name, furtherPath)
        if self.engine is not None:
            unproxied = removeSecurityProxy(provider)
            addTALNamespaceData(unproxied, self.engine)
        return provider


def queryContentProvider(context, request, view, name):
    provider = queryMultiAdapter(
        (context, request, view),
        interfaces.IContentProvider,
        name)
    if provider is not None:
        return provider

    zope_provider = queryMultiAdapter(
        (context, request, view),
        zope.contentprovider.interfaces.IContentProvider,
        name)
    if zope_provider is None:
        return None
    provider = interfaces.IContentProvider(provider, None)
    return provider


@adapter(interfaces.IContentProvider)
@implementer(ICrowd)
def providerViewCrowd(provider):
    return getAdapter(provider.context, ICrowd, name="schooltool.view")


@adapter(interfaces.IContentProvider)
@implementer(ICrowd)
def providerEditCrowd(provider):
    return getAdapter(provider.context, ICrowd, name="schooltool.edit")


@adapter(interfaces.IContentProvider)
@implementer(ICrowd)
def providerManageAppCrowd(provider):
    return getAdapter(provider.context, ICrowd, name="zope.ManageApplication")



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.sampledata.generator
"""
import unittest
import doctest
from pprint import pprint

from zope.app.testing import setup
from zope.publisher.browser import TestRequest
from zope.component import provideUtility
from schooltool.testing.setup import setUpSchoolToolSite
from schooltool.testing.setup import setUpApplicationPreferences
import schooltool.app.browser.testing

def setUp(test):
    setup.placefulSetUp()


def tearDown(test):
    setup.placefulTearDown()


def doctest_SampleDataView_update():
    """Tests for SampleDataView.update method

        >>> app = setUpSchoolToolSite()

    When seed is not provided in the request, the seed is set to
    'SchoolTool', so that sample data is reproducible:

        >>> from schooltool.sampledata.browser import SampleDataView
        >>> request = TestRequest()
        >>> view = SampleDataView(app, request)
        >>> view.update()
        >>> view.seed
        'SchoolTool'

    Otherwise update picks up the seed value:

        >>> request = TestRequest(form={'seed': 'random-data-123'})
        >>> view = SampleDataView(app, request)
        >>> view.update()
        >>> view.seed
        'random-data-123'

    If the seed is an empty string, the seed value is None, so that
    the random number generators initialize from current time.

        >>> request = TestRequest(form={'seed': ''})
        >>> view = SampleDataView(app, request)
        >>> view.update()
        >>> view.seed

    Now, let's set up some stub sample data plugins.

        >>> from schooltool.sampledata.tests.test_generator import DummyPlugin
        >>> from schooltool.sampledata.interfaces import ISampleDataPlugin
        >>> p1 = DummyPlugin("work.day", ())
        >>> p2 = DummyPlugin("play", ("work.day", ))
        >>> provideUtility(p1, ISampleDataPlugin, 'work.day')
        >>> provideUtility(p2, ISampleDataPlugin, 'play')

    If we fill in the seed and press the submit button, we get sample
    data plugins called.

        >>> DummyPlugin.log = []
        >>> request = TestRequest(form={'seed': 'data', 'SUBMIT': 'Generate',
        ...                             'plugin.work.day': 'checked',
        ...                             'plugin.play': 'checked'})
        >>> view = SampleDataView(app, request)
        >>> view.update()
        >>> pprint(DummyPlugin.log)
        [('work.day',
          <schooltool.app.app.SchoolToolApplication object at ...>,
          'data'),
         ('play',
          <schooltool.app.app.SchoolToolApplication object at ...>,
          'data')]

    When the work is done the times attribute is set on the view:

        >>> sorted(view.times)
        ['play', 'work.day']
        >>> for item, time in view.times.items():
        ...     assert time < 0.1   # 100 ms ought to be enough to do nothing
        >>> view.work_done
        True

    If we press the cancel button, we also get redirected to the main page:

        >>> request = TestRequest(form={'seed': 'data', 'CANCEL': 'Cancel'})
        >>> view = SampleDataView(app, request)
        >>> view.update()
        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('location')
        'http://127.0.0.1'

    Clean up:

        >>> DummyPlugin.log = []
    """

def doctest_SampleDataView__call__():
    """Tests for SampleDataView rendering

    We'll need a massive browser views setup here.

        >>> schooltool.app.browser.testing.setUp()
        >>> setUpApplicationPreferences()

    Let's set up some plugins so they are called:

        >>> from schooltool.sampledata.tests.test_generator import DummyPlugin
        >>> from schooltool.sampledata.interfaces import ISampleDataPlugin
        >>> p1 = DummyPlugin("work.day", ())
        >>> p2 = DummyPlugin("play", ("work.day", ))
        >>> provideUtility(p1, ISampleDataPlugin, 'work.day')
        >>> provideUtility(p2, ISampleDataPlugin, 'play')

    Let's create an application object and a view:

        >>> from schooltool.sampledata.browser import SampleDataView
        >>> app = setUpSchoolToolSite()
        >>> request = TestRequest(form={'seed': 'test'})
        >>> view = SampleDataView(app, request)

    Let's render the view.  The default value of the seed is displayed there.

        >>> print view()
        <BLANKLINE>
        ...
        <h1>Sample Data Generation</h1>
        ...
        <form method="POST" class="standalone"
              action="http://127.0.0.1">
          <div class="row">
            <div class="label">
               <label>Random seed</label>
            </div>
            <div class="field">
               <input name="seed" value="test" />
            </div>
          </div>
          <fieldset>
            <legend>Plugins</legend>
            <div class="item_list">
            <div>
              <label>
                <input type="checkbox" name="plugin.play" />
                play
              </label>
            </div>
            <div>
              <label>
                <input type="checkbox" name="plugin.work.day" />
                work.day
              </label>
            </div>
            <div class="controls">
              <input type="submit" class="button-ok"
                     name="DEPENDENCIES" value="View Dependencies" />
              <input type="submit" class="button-ok" name="CLEAR"
                     value="Clear" />
            </div>
          </div></fieldset>
          <div class="controls">
            <input type="submit" class="button-ok" name="SUBMIT"
                   value="Generate" />
            <input type="submit" class="button-cancel" name="CANCEL"
                   value="Cancel" />
          </div>
        </form>
        ...

    Let's render the view.  The default value of the seed is displayed there.

        >>> request = TestRequest(form={'seed': 'test', 'SUBMIT': 'Generate',
        ...                             'plugin.work.day': 'checked',
        ...                             'plugin.play': 'checked'})
        >>> view = SampleDataView(app, request)
        >>> print view()
        <BLANKLINE>
        ...
        <p>Sample data generated.  Below is the list of plugins executed along
        with the CPU time it took to run them</p>
        <BLANKLINE>
        <table>
          <tr>
            <th>Plugin Name</th>
            <th>CPU time used (seconds)</th>
          </tr>
          <tr>
            <td>play</td>
            <td>...</td>
          </tr>
          <tr>
            <td>work.day</td>
            <td>...</td>
          </tr>
        </table>
        ...
    """

def doctest_SampleDataView_work_done():
    """Tests for SampleDataView.work_done propery.

        >>> from schooltool.sampledata.browser import SampleDataView
        >>> request = TestRequest()
        >>> view = SampleDataView(object(), request)

    Before any processing is done, work_done is False:

        >>> view.work_done
        False

    When the update sets the times attribute, it becomes True:

        >>> view.times = {}
        >>> view.work_done
        True

    """

def test_suite():
    return unittest.TestSuite([
        doctest.DocTestSuite(setUp=setUp,
                             tearDown=tearDown,
                             optionflags=doctest.ELLIPSIS
                             |doctest.NORMALIZE_WHITESPACE
                             |doctest.REPORT_NDIFF),
        ])

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.generations.evolve38
"""
import unittest
import doctest

from zope.app.testing import setup
from zope.component import provideHandler, provideAdapter
from zope.component.hooks import getSite, setSite
from zope.interface import implements
from zope.lifecycleevent.interfaces import IObjectModifiedEvent
from zope.site.folder import Folder

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.generations.tests import ContextStub
from schooltool.timetable.app import SCHEDULES_KEY
from schooltool.timetable.app import TIMETABLES_KEY
from schooltool.timetable.app import SchoolToolSchedules
from schooltool.timetable.schedule import Schedule, ScheduleContainer
from schooltool.timetable.timetable import Timetable, SelectedPeriodsSchedule


class AppStub(Folder):
    implements(ISchoolToolApplication)

    def __init__(self):
        super(AppStub, self).__init__()


def doctest_evolve38():
    r"""Test evolution to generation 38.

        >>> context = ContextStub()
        >>> context.root_folder['app'] = app = AppStub()
        >>> manager = setup.createSiteManager(app)

        >>> schedules = app[SCHEDULES_KEY] = SchoolToolSchedules()
        >>> timetables = app[TIMETABLES_KEY] = SchoolToolSchedules()

        >>> events = []
        >>> def log_event(event):
        ...     events.append('%s: %s / %s' %
        ...         (event.__class__.__name__,
        ...          event.object.__parent__.__name__,
        ...          event.object.__name__))
        >>> provideHandler(log_event, [IObjectModifiedEvent])

        >>> provideAdapter(lambda ignore:app, (None,), ISchoolToolApplication)


    Let's build two timetables.

        >>> tt1 = timetables['1'] = Timetable('starts', 'ends')
        >>> tt2 = timetables['2'] = Timetable('starts', 'ends')

    Add schedule container that does not use them.

        >>> schedules['0'] = ScheduleContainer()
        >>> schedules['0']['1'] = Schedule('starts', 'ends')

    And some containers that do use the timetables.

        >>> schedules['1'] = ScheduleContainer()
        >>> schedules['1']['1'] = SelectedPeriodsSchedule(tt1, 'starts', 'ends')
        >>> schedules['2'] = ScheduleContainer()
        >>> schedules['2']['1'] = SelectedPeriodsSchedule(tt2, 'starts', 'ends')
        >>> schedules['2']['2'] = SelectedPeriodsSchedule(tt2, 'starts', 'ends')

        >>> events[:] = []

    Let's evolve now.

        >>> from schooltool.generations.evolve38 import evolve
        >>> evolve(context)

    ObjectModifiedEvent was fired for each of schedule containers
    that have schedules using any of app's timetables.

        >>> print '\n'.join(events)
        ObjectModifiedEvent: schooltool.timetable.schedules / 1
        ObjectModifiedEvent: schooltool.timetable.schedules / 2

    Site was restored after evolution.

        >>> print getSite()
        None

    """


def setUp(test):
    setup.placelessSetUp()
    setup.setUpTraversal()
    setSite()

def tearDown(test):
    setSite()
    setup.placelessTearDown()


def test_suite():
    optionflags = (doctest.ELLIPSIS |
                   doctest.NORMALIZE_WHITESPACE |
                   doctest.REPORT_ONLY_FIRST_FAILURE)
    return doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                optionflags=optionflags)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.app.
"""
import unittest
import doctest

from schooltool.skin.flourish import sorting


def doctest_scc_tarjan():
    """Tests for Tarjan's strongly connected components.

    First, we need items to sort.

        >>> items = ['fire', 'water', 'cheese', 'goat', 'cabbage']

    Items are used as a priority queue; if there are no dependencies,
    queue is left unchanged.

        >>> sorting.scc_tarjan(items, {})
        [('fire',), ('water',), ('cheese',), ('goat',), ('cabbage',)]

    Say we have some dependencies:

        >>> graph = {'cheese': ['goat']}
        >>> sorting.scc_tarjan(items, graph)
        [('fire',), ('water',), ('goat',), ('cheese',), ('cabbage',)]

        >>> graph = {'cheese': ['goat'], 'goat': ['cabbage']}
        >>> sorting.scc_tarjan(items, graph)
        [('fire',), ('water',), ('cabbage',), ('goat',), ('cheese',)]

    Tarjan's algorithm is mainly used to detect cyclic dependencies.

        >>> graph = {'goat': ['cabbage'], 'cabbage': ['water'],
        ...          'fire': ['cabbage'], 'water': ['goat']}
        >>> sorting.scc_tarjan(items, graph)
        [('cabbage', 'water', 'goat'), ('fire',), ('cheese',)]

    """


def doctest_topological_sort():
    """Tests for a simple implementation of a topological sort.

        >>> items = ['1', '2', '3', '4', '5', '6']
        >>> graph = {'1': ['3'], '5': ['1'], '6': ['3']}

    Graph is a dict of dependencies, 'x' comes after ['y1', 'y2']

        >>> sorting.topological_sort(items, graph)
        ['2', '3', '1', '5', '6', '4']

    topological_sort can also accept a dict of reverse dependencies,
    'x' comes before ['y1', 'y2']

        >>> sorting.topological_sort(items, graph, reverse=True)
        ['2', '5', '1', '6', '3', '4']

    """


def doctest_dependency_graph():
    """Tests for dependency_graph builder.

        The builder takes dicts of revers and direct dependencies and
        makes a combined direct dependecy graph.

        >>> before = {'1': ['3'], '5': ['3'], '4': ['1']}
        >>> after = {'3': ['4', '5', '6']}

        >>> graph = sorting.dependency_graph(before, after)
        >>> for n, d in sorted(graph.items()):
        ...     print n, ':', sorted(d)
        1 : ['4']
        3 : ['1', '4', '5', '6']

    """


def doctest_dependency_sort():
    """Tests for dependency_sort

    dependency_sort sorts items given direct and reverse dependencies.

        >>> items = ['under', 'pants', 'shirt',
        ...          'shoes', 'socks', 'watch', 'tie']

        >>> after = {
        ...     'pants': ['under', 'shirt'],
        ...     }

        >>> before = {
        ...     'pants': ['shoes'],
        ...     'shirt': ['tie'],
        ...     'socks': ['shoes'],
        ...     }

        >>> sorting.dependency_sort(items, before, after)
        ['under', 'socks', 'shirt', 'pants', 'shoes', 'watch', 'tie']


    Usually dependencies can be sorted in many valid ways.  This implementation
    tries to keep 'before' and 'after' items close to their dependencies, at
    the same time maintaining original order.

    Note that '6' after '4' gives different results than '4' before '6':

        >>> items = ['1', '2', '3', '4', '5', '6', '7']

        >>> after = {'6': ['4']}
        >>> before = {'4': ['2']}
        >>> sorting.dependency_sort(items, before, after)
        ['1', '4', '6', '2', '3', '5', '7']


        >>> after = {'2': ['4']}
        >>> before = {'4': ['6']}
        >>> sorting.dependency_sort(items, before, after)
        ['1', '3', '5', '4', '2', '6', '7']


    Here are some more loose tests:

        >>> after = {'3': ['1']}
        >>> before = {'1': ['2'], '5': ['2'],}
        >>> sorting.dependency_sort(items, before, after)
        ['1', '3', '5', '2', '4', '6', '7']

        >>> after = {'2': ['5']}
        >>> before = {'3': ['2'], '4': ['2']}
        >>> sorting.dependency_sort(items, before, after)
        ['1', '3', '4', '5', '2', '6', '7']

        >>> after = {'2': ['4'], '4': ['3']}
        >>> before = {'3': ['6']}
        >>> sorting.dependency_sort(items, before, after)
        ['1', '5', '3', '4', '2', '6', '7']

        >>> after = {'2': ['3'], '3': ['6']}
        >>> before = {'5': ['4']}
        >>> sorting.dependency_sort(items, before, after)
        ['1', '5', '4', '6', '3', '2', '7']


    """


def test_suite():
    optionflags = (doctest.ELLIPSIS |
                   doctest.NORMALIZE_WHITESPACE |
                   doctest.REPORT_NDIFF)

    return unittest.TestSuite([
                doctest.DocTestSuite(optionflags=optionflags),
           ])


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for schooltool content provider machinery.
"""
import unittest
import doctest

from zope.tales.tales import CompilerError

from schooltool.app.browser.testing import setUp, tearDown


class PythonExp(object):

    def __init__(self, exp):
        self.exp = exp

    def __call__(self, econtext):
        return eval(self.exp)


class TalVariableExp(object):

    def __init__(self, exp):
        self.exp = exp

    def __call__(self, econtext):
        return econtext.vars[self.exp.strip()]


class TalContext(object):

    def __init__(self, vars=()):
        self.vars = dict(vars)


class TestTalEngine(object):

    def compile(self, expr):
        if expr.startswith('python:'):
            return PythonExp(expr[7:])
        return TalVariableExp(expr)

    def getCompilerError(self):
        return CompilerError


def doctest_JSONEncoder():
    """Tests for JSONEncoder.

        >>> from schooltool.skin.flourish.tal import JSONEncoder

    JSONEncoder is extended version of python's json.JSONEncoder

        >>> import json

        >>> issubclass(JSONEncoder, json.JSONEncoder)
        True

    We can encode some of the standart python types to JSON.

        >>> encoder = JSONEncoder()

        >>> data = {'hello': 'world', 'foo': ('bar', 2, 'baz')}
        >>> encoder.encode(data)
        '{"foo": ["bar", 2, "baz"], "hello": "world"}'

    In addition, we can also encode generators.

        >>> def countme(x):
        ...     for y in range(x):
        ...         yield y

        >>> encoder.encode(countme(5))
        '[0, 1, 2, 3, 4]'

        >>> encoder.encode({'numbers': countme(3)})
        '{"numbers": [0, 1, 2]}'

        >>> print encoder.encode({'Pe/ter': 'Ahmaleethaxor</script>!'})
        {"Pe\\/ter": "Ahmaleethaxor<\\/script>!"}

    """


def doctest_JSONExpression():
    """Tests for JSONExpression.

        >>> from schooltool.skin.flourish.tal import JSONExpression

        >>> engine = TestTalEngine()

    This expression encodes a result of a TAL expression to JSON.

        >>> exp = JSONExpression(
        ...     'json', 'python:{"hello": ("world", )}', engine)

        >>> exp(TalContext())
        '{"hello":["world"]}'

    __str__ and __repr__ looks similar to the rest of TAL expressions.

        >>> print exp
        json expression ('python:{"hello": ("world", )}')

        >>> print repr(exp)
        <JSONExpression json:python:{"hello": ("world", )}>

    """


def doctest_ScriptLocalExpression():
    """Tests for ScriptLocalExpression.

        >>> from schooltool.skin.flourish.tal import ScriptLocalExpression

        >>> engine = TestTalEngine()

        >>> exp = ScriptLocalExpression('scriptlocal',
        ...     'foo python:2*2, 3*5', engine)

        >>> print exp(TalContext())
        <script>$.extend(true,ST.local,{"foo":[4,15]});</script>

        >>> exp = ScriptLocalExpression('scriptlocal',
        ...     'nine python: 3*3;'
        ...     'ten python: 2*5;', engine)

        >>> print exp(TalContext())
        <script>$.extend(true,ST.local,{"nine":9,"ten":10});</script>

        >>> context = TalContext({'name': 'Bob', 'age': 10})

        >>> exp = ScriptLocalExpression('scriptlocal',
        ...     'student name', engine)

        >>> print exp(context)
        <script>$.extend(true,ST.local,{"student":"Bob"});</script>

        >>> exp = ScriptLocalExpression('scriptlocal',
        ...     'name; age', engine)

        >>> print exp(context)
        <script>$.extend(true,ST.local,{"age":10,"name":"Bob"});</script>

    """


def doctest_ScriptLocalExpression_syntax():
    """Tests for ScriptLocalExpression syntax errors.

        >>> from schooltool.skin.flourish.tal import ScriptLocalExpression

        >>> engine = TestTalEngine()

    Bad syntax raises CompilerError:

        >>> ScriptLocalExpression('scriptlocal', 'mumbo+jumbo hey', engine)
        Traceback (most recent call last):
        ...
        CompilerError: Invalid script local expression 'mumbo+jumbo hey'
                       (expected: "js_var_name tales_expression").

        >>> ScriptLocalExpression('scriptlocal', ' good; b?a/d', engine)
        Traceback (most recent call last):
        ...
        CompilerError: Invalid script local expression 'b?a/d'
                       (expected: "js_var_name tales_expression").

    As well as duplicate variable names do:

        >>> ScriptLocalExpression('scriptlocal', 'tony foo; tony bar', engine)
        Traceback (most recent call last):
        ...
        CompilerError: Duplicate javascript variable name 'tony' in
                       'tony foo; tony bar'.

    Variable names cannot start with digits.

        >>> ScriptLocalExpression('scriptlocal', '7bad', engine)
        Traceback (most recent call last):
        ...
        CompilerError: Invalid script local expression '7bad'
                       (expected: "js_var_name tales_expression").

    When tal expression is not specified (it's assumed to be identical
    to the variable name), variable names must be valid TAL variables.

        >>> ScriptLocalExpression('scriptlocal', '$bogus', engine)
        Traceback (most recent call last):
        ...
        CompilerError: When only javascript variable name is specified
                       ('$bogus') it must be a valid TAL variable.

    Names starting with underscore are valid:

        >>> ScriptLocalExpression('scriptlocal', '_hello_world', engine)
        <ScriptLocalExpression ['_hello_world']>

    """


def test_suite():
    optionflags = (doctest.ELLIPSIS | doctest.REPORT_NDIFF
                   | doctest.NORMALIZE_WHITESPACE)
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                       optionflags=optionflags))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Email interfaces

"""

from zope.container.constraints import contains, containers
from zope.location.interfaces import IContained
from zope.container.interfaces import IReadContainer, IWriteContainer
from zope.interface import Interface
from zope.location.interfaces import ILocation
from zope.schema import Bool, Datetime, Dict
from zope.schema import TextLine, List, Text, Int, Password

from schooltool.common import SchoolToolMessage as _


class IEmail(Interface):
    """An email"""

    from_address = TextLine(
        title=_(u'From'),
        description=_(u'The sender address'))

    to_addresses = List(
        title=_(u'To'),
        description=_(u'Recipient addresses'),
        value_type=TextLine(title=_(u'Recipient address')),
        min_length=1)

    body = Text(
        title=_(u'Body'),
        description=_(u'Body of the message'))

    subject = TextLine(
        title=_(u'Subject'),
        description=_(u'Subject of the message'),
        required=False)

    status_code = TextLine(
        title=_(u'Status'),
        description=_(u'Code for the status of the message'),
        required=False)

    status_parameters = Dict(
        title=_(u'Status Parameters'),
        description=_(u'Parameters for the status of the message'),
        key_type=TextLine(title=_(u'Parameter key')),
        value_type=TextLine(title=_(u'Parameter value')),
        required=False)

    time_created = Datetime(
        title=_(u'Created on'),
        description=_(u'Date and time when the message was created'))

    time_sent = Datetime(
        title=_(u'Sent on'),
        description=_(u'Date and time when the message was sent'),
        required=False)


class IWriteEmailContainer(IWriteContainer):
    """Write interface for the IEmailContainer"""


class IReadEmailContainer(IReadContainer, ILocation):
    """Read interface for the IEmailContainer"""

    enabled = Bool(
        title=_('Enable'),
        description=_('Mark to enable the service.'),
        default=False)

    hostname = TextLine(
        title=_('Hostname'),
        description=_('SMTP server hostname. Required if the service is enabled.'),
        required=False)

    port = Int(
        title=_('Port'),
        description=_('Port of the SMTP service. Using 25 by default.'),
        min=0,
        default=25,
        required=False)

    username = TextLine(
        title=_('Username'),
        description=_('Username used for optional SMTP authentication.'),
        required=False)

    password = Password(
        title=_('New Password'),
        description=_('The password to authenticate to the SMTP server.'),
        required=False)

    tls = Bool(
        title=_('TLS'),
        description=_('Use TLS connection?'),
        default=False,
        required=False)


class IEmailContainer(IWriteEmailContainer, IReadEmailContainer):
    """A container for IEmail objects"""

    contains(IEmail)


class IEmailContained(IEmail, IContained):
    """An IEmail object that is contained in a IEmailContainer"""

    containers(IEmailContainer)


class IEmailUtility(Interface):
    """An utility to send IEmail objects"""

    def send(email):
        """Sends an email message.

        `email` is a IEmail object.

        Returns True if the email is sent successfully to all its
        recipients. False otherwise.

        If this method returns False, the `email` object is added to
        the IEmailContainer if it is not there.

        """

    def enabled():
        """Checks if the email service is enabled.

        Returns True if the service is enabled. False otherwise.

        """



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.timetable.timetable
"""
import doctest
import pytz
import unittest
from pprint import pprint
from datetime import datetime, date, time, timedelta

from zope.app.testing import setup
from zope.intid.interfaces import IIntIds
from zope.component import provideUtility, provideAdapter
from test_schedule import ScheduleStub

from schooltool.timetable.calendar import ImmutableScheduleCalendar
from schooltool.timetable.interfaces import IHaveSchedule


class ImmutableScheduleCalendarForTest(ImmutableScheduleCalendar):

    def makeGUID(self, date, period, int_ids=None):
        return u'%s.%s' % (
            date.isoformat(),
            period and period.title or '?',
            )


def print_events(cal):
    for event in cal:
        print event.title, 'on', event.dtstart.strftime('%Y-%m-%d %H:%M %Z')


def test_ImmutableScheduleCalendar():
    """Tests for ImmutableScheduleCalendar.

    Let's take a simple schedule that spans a dst shift in some timezone.

        >>> schedule = ScheduleStub(timezone='Europe/Vilnius')

        >>> pprint(list(schedule.iterMeetings(schedule.first, schedule.last)))
        [<Meeting on 2011-10-29 00:05 EEST>,
         <Meeting on 2011-10-29 05:00 EEST>,
         <Meeting on 2011-10-29 23:55 EEST>,
         <Meeting on 2011-10-30 00:05 EEST>,
         <Meeting on 2011-10-30 05:00 EET>,
         <Meeting on 2011-10-30 23:55 EET>,
         <Meeting on 2011-10-31 00:05 EET>,
         <Meeting on 2011-10-31 05:00 EET>,
         <Meeting on 2011-10-31 23:55 EET>]

    Let's build a calendar for, say, math class:

        >>> class Math(object):
        ...     ___init__ = lambda self, schedule: None
        ...     title = 'Math'
        >>> provideAdapter(lambda s: Math, (ScheduleStub, ), IHaveSchedule)

        >>> cal = ImmutableScheduleCalendarForTest(schedule)

    Calendar events are always localized to UTC, but we can still see
    the correct compensation for dst shift:

        >>> print_events(cal)
        Math on 2011-10-28 21:05 UTC
        Math on 2011-10-29 02:00 UTC
        Math on 2011-10-29 20:55 UTC
        Math on 2011-10-29 21:05 UTC
        Math on 2011-10-30 03:00 UTC
        Math on 2011-10-30 21:55 UTC
        Math on 2011-10-30 22:05 UTC
        Math on 2011-10-31 03:00 UTC
        Math on 2011-10-31 21:55 UTC

    """


def setUp(test=None):
    setup.placelessSetUp()
    provideUtility(object(), IIntIds)


def tearDown(test=None):
    setup.placelessTearDown()


def test_suite():
    optionflags = (doctest.ELLIPSIS | doctest.REPORT_NDIFF |
                   doctest.NORMALIZE_WHITESPACE)
    return unittest.TestSuite([
            doctest.DocTestSuite(optionflags=optionflags,
                                 setUp=setUp, tearDown=tearDown),
           ])



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Browser views for notes.
"""

from zope.publisher.browser import BrowserView
from zope.app.form.browser.add import AddView
from zope.traversing.browser.absoluteurl import absoluteURL

from schooltool.note.interfaces import IHaveNotes, INotes, INote
from schooltool.note.note import Note


class NotesView(BrowserView):

    # __used_for__ = ?

    def __init__(self, context, request):
        BrowserView.__init__(self, context, request)
        notes = INotes(context)
        self.notes = [note for note in notes
                      if note.privacy == 'public'
                         or note.owner == request.principal.id]

        if 'DELETE_NOTE' in request:
            notes.remove(request['uid'])


class NoteAddView(AddView):
    """A view for adding a note."""

    __used_for__ = IHaveNotes

    # Form error message for the page template
    error = None

    # Override some fields of AddView
    schema = INote
    _factory = Note
    _arguments = ['title', 'body', 'privacy']
    _keyword_arguments = []
    _set_before_add = []
    _set_after_add = []

    def create(self, title, body, privacy):
        owner = self.request.principal.id
        note = self._factory(title=title, body=body, privacy=privacy,
                             owner=owner)
        return note

    def add(self, note):
        """Add `note` to the object."""
        notes = INotes(self.context)
        notes.add(note)
        return note

    def update(self):
        if 'CANCEL' in self.request:
            url = absoluteURL(self.context, self.request)
            self.request.response.redirect(url)
        return AddView.update(self)

    def nextURL(self):
        """See zope.browser.interfaces.IAdding"""
        return absoluteURL(self.context, self.request)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Implementation of relationships.

Relationships are represented as collections of links.  A link defines one
half of a relationship.  The storage of links on an object is determined by
an IRelationshipLinks adapter.  There is a default adapter registered for
all IAnnotatable objects that uses Zope 3 annotations.
"""

from BTrees.OOBTree import OOBTree
from persistent import Persistent
from persistent.list import PersistentList
from zope.container.contained import Contained
from zope.interface import implements
import zope.event

from schooltool.relationship.interfaces import IRelationshipLinks
from schooltool.relationship.interfaces import IRelationshipInfo
from schooltool.relationship.interfaces import IRelationshipLink
from schooltool.relationship.interfaces import IRelationshipProperty
from schooltool.relationship.interfaces import IBeforeRelationshipEvent
from schooltool.relationship.interfaces import IRelationshipAddedEvent
from schooltool.relationship.interfaces import IBeforeRemovingRelationshipEvent
from schooltool.relationship.interfaces import IRelationshipRemovedEvent
from schooltool.relationship.interfaces import DuplicateRelationship
from schooltool.relationship.interfaces import NoSuchRelationship
from schooltool.relationship.interfaces import IRelationshipSchema


def relate(rel_type, (a, role_of_a), (b, role_of_b), extra_info=None):
    """Establish a relationship between objects `a` and `b`."""
    for link in IRelationshipLinks(a):
        if (link.target is b and link.role == role_of_b
            and link.rel_type == rel_type):
            raise DuplicateRelationship
    zope.event.notify(BeforeRelationshipEvent(rel_type,
                                              (a, role_of_a),
                                              (b, role_of_b),
                                              extra_info))
    IRelationshipLinks(a).add(Link(role_of_a, b, role_of_b, rel_type,
                                   extra_info))
    IRelationshipLinks(b).add(Link(role_of_b, a, role_of_a, rel_type,
                                   extra_info))
    zope.event.notify(RelationshipAddedEvent(rel_type,
                                             (a, role_of_a),
                                             (b, role_of_b),
                                             extra_info))


def unrelate(rel_type, (a, role_of_a), (b, role_of_b)):
    """Break a relationship between objects `a` and `b`."""
    links_of_a = IRelationshipLinks(a)
    links_of_b = IRelationshipLinks(b)
    try:
        link_a_to_b = links_of_a.find(role_of_a, b, role_of_b, rel_type)
    except ValueError:
        raise NoSuchRelationship
    extra_info = link_a_to_b.extra_info
    zope.event.notify(BeforeRemovingRelationshipEvent(rel_type,
                                                      (a, role_of_a),
                                                      (b, role_of_b),
                                                      extra_info))
    links_of_a.remove(link_a_to_b)
    # If links_of_b.find raises a ValueError, our data structures are out of
    # sync.
    link_b_to_a = links_of_b.find(role_of_b, a, role_of_a, rel_type)
    links_of_b.remove(link_b_to_a)
    zope.event.notify(RelationshipRemovedEvent(rel_type,
                                               (a, role_of_a),
                                               (b, role_of_b),
                                               extra_info))


def unrelateAll(obj):
    """Break all relationships of `obj`.

    Note that this operation is not atomic: if an event subscriber catches
    a BeforeRemovingRelationshipEvent and vetoes the operation, some
    relationships may have been removed, while others may still be there.
    """
    links_of_a = IRelationshipLinks(obj)
    relationships = [(link.rel_type, (obj, link.my_role),
                                     (link.target, link.role))
                     for link in links_of_a]
    for args in relationships:
        try:
            unrelate(*args)
        except NoSuchRelationship:
            pass # it was a loop, so we tried to delete it twice
    return


class RelationshipEvent(object):
    """Base class for relationship events.

        >>> event = RelationshipEvent('Membership',
        ...                           ('a', 'Member'), ('b', 'Group'),
        ...                           None)
        >>> event['Member']
        'a'
        >>> event['Group']
        'b'
        >>> event['Bogus']
        Traceback (most recent call last):
          ...
        KeyError: 'Bogus'

    """

    def __init__(self, rel_type, (a, role_of_a), (b, role_of_b), extra_info):
        self.rel_type = rel_type
        self.participant1 = a
        self.role1 = role_of_a
        self.participant2 = b
        self.role2 = role_of_b
        self.extra_info = extra_info

    def __getitem__(self, role):
        """Return the participant with a given role."""
        if role == self.role1:
            return self.participant1
        if role == self.role2:
            return self.participant2
        raise KeyError(role)

    def match(self, schema):
        if self.rel_type != schema.rel_type:
            return None
        schema_roles = tuple(schema.roles.values())
        if ((self.role1, self.role2) != schema_roles and
            (self.role2, self.role1) != schema_roles):
            return None
        return RelationshipMatch(self, schema)


class RelationshipMatch(object):

    def __init__(self, event, schema):
        self._event = event
        self._schema = schema
        self.extra_info = event.extra_info
        for name, role in schema.roles.items():
            if role == event.role1:
                setattr(self, name, event.participant1)
            elif role == event.role2:
                setattr(self, name, event.participant2)


class BeforeRelationshipEvent(RelationshipEvent):
    """A relationship is about to be established.

        >>> from zope.interface.verify import verifyObject
        >>> event = BeforeRelationshipEvent('example:Membership',
        ...                                 ('a', 'example:Member'),
        ...                                 ('letters', 'example:Group'),
        ...                                 None)
        >>> verifyObject(IBeforeRelationshipEvent, event)
        True

    """

    implements(IBeforeRelationshipEvent)


class RelationshipAddedEvent(RelationshipEvent):
    """A relationship has been established.

        >>> from zope.interface.verify import verifyObject
        >>> event = RelationshipAddedEvent('example:Membership',
        ...                                ('a', 'example:Member'),
        ...                                ('letters', 'example:Group'),
        ...                                None)
        >>> verifyObject(IRelationshipAddedEvent, event)
        True

    """

    implements(IRelationshipAddedEvent)


class BeforeRemovingRelationshipEvent(RelationshipEvent):
    """A relationship is about to be broken.

        >>> from zope.interface.verify import verifyObject
        >>> event = BeforeRemovingRelationshipEvent('example:Membership',
        ...                 ('a', 'example:Member'),
        ...                 ('letters', 'example:Group'),
        ...                 None)
        >>> verifyObject(IBeforeRemovingRelationshipEvent, event)
        True

    """

    implements(IBeforeRemovingRelationshipEvent)


class RelationshipRemovedEvent(RelationshipEvent):
    """A relationship has been broken.

        >>> from zope.interface.verify import verifyObject
        >>> event = RelationshipRemovedEvent('example:Membership',
        ...                                  ('a', 'example:Member'),
        ...                                  ('letters', 'example:Group'),
        ...                                  None)
        >>> verifyObject(IRelationshipRemovedEvent, event)
        True

    """

    implements(IRelationshipRemovedEvent)


def getRelatedObjects(obj, role, rel_type=None):
    """Return all objects related to `obj` with a given role."""
    return IRelationshipLinks(obj).getTargetsByRole(role, rel_type)


class RelationshipSchema(object):
    """Relationship schema.

    Boring doctest setup:

        >>> from schooltool.relationship.tests import setUp, tearDown
        >>> from schooltool.relationship.tests import SomeObject
        >>> setUp()
        >>> a = SomeObject('a')
        >>> b = SomeObject('b')

    Relationship schemas are syntactic sugar.  If you define a relationship
    schema like this:

        >>> from schooltool.relationship.tests import URIStub
        >>> URIMembership = URIStub('example:Membership')
        >>> URIMember = URIStub('example:Member')
        >>> URIGroup = URIStub('example:Group')
        >>> Membership = RelationshipSchema(URIMembership,
        ...                     member=URIMember, group=URIGroup)

    Then you can create, query and break relationships by writing

        >>> Membership(member=a, group=b)
        >>> Membership.query(group=b)
        [a]
        >>> Membership.unlink(member=a, group=b)

    instead of having to explicitly say

        >>> relate(URIMembership, (a, URIMember), (b, URIGroup))
        >>> getRelatedObjects(b, URIMember, rel_type=URIMembership)
        [a]
        >>> unrelate(URIMembership, (a, URIMember), (b, URIGroup))

    That's it.

        >>> tearDown()

    """

    implements(IRelationshipSchema)

    def __init__(self, rel_type, **roles):
        if len(roles) != 2:
            raise TypeError("A relationship must have exactly two ends.")
        self.rel_type = rel_type
        self.roles = roles

    def __call__(self, **parties):
        """Establish a relationship."""
        self._doit(relate, **parties)

    def unlink(self, **parties):
        """Break a relationship."""
        self._doit(unrelate, **parties)

    def query(self, **party):
        """Retrieve relationship targets."""
        if len(party) != 1:
            raise TypeError("A single party must be provided.")
        roles = list(self.roles.keys())
        my_role_key = party.keys()[0]
        roles.remove(my_role_key)
        other_role = self.roles[roles[0]]
        obj = party.values()[0]
        return getRelatedObjects(obj, other_role, rel_type=self.rel_type)

    def _doit(self, fn, **parties):
        """Extract and validate parties from keyword arguments and call fn."""
        (name_of_a, role_of_a), (name_of_b, role_of_b) = self.roles.items()
        try:
            a = parties.pop(name_of_a)
        except KeyError:
            raise TypeError('Missing a %r keyword argument.' % name_of_a)
        try:
            b = parties.pop(name_of_b)
        except KeyError:
            raise TypeError('Missing a %r keyword argument.' % name_of_b)
        if parties:
            raise TypeError("Too many keyword arguments.")
        fn(self.rel_type, (a, role_of_a), (b, role_of_b))


class RelationshipProperty(object):
    """Relationship property.

        >>> from zope.annotation.interfaces import IAttributeAnnotatable
        >>> from schooltool.relationship.tests import setUp, tearDown
        >>> setUp()
        >>> from schooltool.relationship.tests import URIStub

    Instead of calling global functions and passing URIs around you can define
    a property on an object and use it to create and query relationships:

        >>> class SomeClass(object):
        ...     implements(IAttributeAnnotatable)
        ...     friends = RelationshipProperty(URIStub('example:Friendship'),
        ...                                    URIStub('example:Friend'),
        ...                                    URIStub('example:Friend'))

    The property is introspectable, although that's not very useful

        >>> SomeClass.friends.rel_type.uri
        'example:Friendship'

        >>> SomeClass.friends.my_role.uri
        'example:Friend'

        >>> SomeClass.friends.other_role.uri
        'example:Friend'

    IRelationshipProperty defines things you can do with a relationship
    property.

        >>> from zope.interface.verify import verifyObject
        >>> someinstance = SomeClass()
        >>> verifyObject(IRelationshipProperty, someinstance.friends)
        True
        >>> tearDown()

    """

    def __init__(self, rel_type, my_role, other_role):
        self.rel_type = rel_type
        self.my_role = my_role
        self.other_role = other_role

    def __get__(self, instance, owner):
        """Bind the property to an instance."""
        if instance is None:
            return self
        else:
            return BoundRelationshipProperty(instance, self.rel_type,
                                             self.my_role, self.other_role)


class BoundRelationshipProperty(object):
    """Relationship property bound to an object."""

    implements(IRelationshipProperty)

    def __init__(self, this, rel_type, my_role, other_role):
        self.this = this
        self.rel_type = rel_type
        self.my_role = my_role
        self.other_role = other_role

    def __nonzero__(self):
        try:
            iter(self).next()
        except StopIteration:
            return False
        else:
            return True

    def __len__(self):
        count = 0
        for i in self:
            count += 1
        return count

    def __iter__(self):
        return iter(getRelatedObjects(self.this, self.other_role,
                                      self.rel_type))

    @property
    def relationships(self):
        linkset = IRelationshipLinks(self.this).getLinksByRole(self.other_role)
        return [RelationshipInfo(self.this, link)
                for link in linkset
                if link.rel_type == self.rel_type]

    def add(self, other, extra_info=None):
        """Establish a relationship between `self.this` and `other`."""
        relate(self.rel_type, (self.this, self.my_role),
                              (other, self.other_role), extra_info)

    def remove(self, other):
        """Unlink a relationship between `self.this` and `other`."""
        unrelate(self.rel_type, (self.this, self.my_role),
                                (other, self.other_role))


class RelationshipInfo(object):
    """Ugly implementation of access to relationship extra information.

    Setup:

        >>> from schooltool.relationship.tests import setUp, tearDown
        >>> from schooltool.relationship.tests import setUpRelationships
        >>> from schooltool.relationship.tests import SomeObject
        >>> from schooltool.relationship.tests import URIStub
        >>> setUp()
        >>> setUpRelationships()

    Say we relate two objects.

        >>> URIMembership = URIStub('example:Membership')
        >>> URIMember = URIStub('example:Member')
        >>> URIGroup = URIStub('example:Group')

        >>> a = SomeObject('a')
        >>> b = SomeObject('b')

        >>> relate(URIMembership, (a, URIMember), (b, URIGroup))

    Two links were created, one for each object.

        >>> link_to_b = list(IRelationshipLinks(a))[0]
        >>> link_to_a = list(IRelationshipLinks(b))[0]

    We will now construct a RelationshipInfo with object and it's link.

        >>> info_a = RelationshipInfo(a, link_to_b)
        >>> info_a.source
        a

        >>> info_a.target
        b

        >>> print info_a.extra_info
        None

    Setting extra_info updates both links.

        >>> info_a.extra_info = 'extra'
        >>> link_to_a.extra_info
        'extra'
        >>> link_to_b.extra_info
        'extra'

    That's it.

        >>> tearDown()

    """

    implements(IRelationshipInfo)

    def __init__(self, this, link_to_other):
        self._this = this
        self._link = link_to_other

    @property
    def source(self):
        return self._this

    @property
    def target(self):
        return self._link.target

    @property
    def extra_info(self):
        return self._link.extra_info

    @extra_info.setter
    def extra_info(self, value):
        this_link = self._link

        links_of_target = IRelationshipLinks(self._link.target)
        # If links_of_target.find raises a ValueError, our data structures are
        # out of sync.
        other_link = links_of_target.find(
            this_link.role, self._this, this_link.my_role, this_link.rel_type)

        this_link.extra_info = value
        other_link.extra_info = value


class Link(Persistent, Contained):
    """One half of a relationship.

    A link is a simple class that holds information about one side of the
    relationship:

        >>> target = object()
        >>> my_role = 'example:Group'
        >>> role = 'example:Member'
        >>> rel_type = 'example:Membership'
        >>> link = Link(my_role, target, role, rel_type)
        >>> link.target is target
        True
        >>> link.my_role
        'example:Group'
        >>> link.role
        'example:Member'
        >>> link.rel_type
        'example:Membership'

    The attributes are documented in IRelationshipLink

        >>> from zope.interface.verify import verifyObject
        >>> verifyObject(IRelationshipLink, link)
        True

    """

    implements(IRelationshipLink)

    def __init__(self, my_role, target, role, rel_type, extra_info=None):
        self.my_role = my_role
        self.target = target
        self.role = role
        self.rel_type = rel_type
        self.extra_info = extra_info


class LinkSet(Persistent, Contained):
    """Set of links.

    This class is used internally to represent relationships.  Initially it
    is empty

        >>> linkset = LinkSet()
        >>> list(linkset)
        []

    You can add new links to it

        >>> from schooltool.relationship.tests import URIStub
        >>> link1 = Link('example:Group', object(), URIStub('example:Member'),
        ...              URIStub('example:Membership'))
        >>> link2 = Link('example:Friend', object(), URIStub('example:Friend'),
        ...              URIStub('example:Friendship'))
        >>> linkset.add(link1)
        >>> linkset.add(link2)

    The links have landed in the cache too:

        >>> expected = {
        ...     'example:Member': [link1],
        ...     'example:Friend': [link2]}
        >>> dict(linkset._byrole.items()) == expected
        True

    Let's zap the cache and call getLinksByRole(), which should restore it:

        >>> del linkset._byrole
        >>> linkset.getLinksByRole(URIStub('something'))
        []

        >>> dict(linkset._byrole.items()) == expected
        True

    Links should get named:

        >>> link1.__name__
        '1'
        >>> link2.__name__
        '2'

    We can access our links through their names:

        >>> linkset['1'] is link1
        True
        >>> linkset['2'] is link2
        True

    And get parents set:

        >>> link1.__parent__ is linkset
        True

    We got them in the container now:

        >>> set(linkset) == set([link1, link2]) # order is not preserved
        True

    You can look for links for a specific relationship

        >>> linkset.find('example:Group',
        ...              link1.target,
        ...              URIStub('example:Member'),
        ...              URIStub('example:Membership')) is link1
        True

    We can't add same link into the container twice:

        >>> linkset.add(link1)                      # doctest: +ELLIPSIS
        Traceback (most recent call last):
          ...
        ValueError: ...

    If find fails, it raises ValueError, just like list.index.

        >>> linkset.find('example:Member', link1.target,
        ...              URIStub('example:Group'),
        ...              URIStub('example:Membership'))      # doctest: +ELLIPSIS
        Traceback (most recent call last):
          ...
        ValueError: ...

    You can remove links

        >>> linkset.remove(link2)
        >>> set(linkset) == set([link1])
        True

    The links are removed from the cache too:

        >>> list(linkset._byrole.keys())
        ['example:Member']

    If you try to remove a link that is not in the set, you will get a
    ValueError.

        >>> linkset.remove(link2)                   # doctest: +ELLIPSIS
        Traceback (most recent call last):
          ...
        ValueError: ...

    You can remove all links

        >>> linkset.clear()
        >>> set(linkset) == set([])
        True

    The cache has been cleared too:

        >>> len(linkset._byrole)
        0

    The class is documented in IRelationshipLinks

        >>> from zope.interface.verify import verifyObject
        >>> verifyObject(IRelationshipLinks, linkset)
        True

    """

    implements(IRelationshipLinks)

    def __init__(self):
        self._links = OOBTree()
        self._byrole = OOBTree() # role -> list of links

    def getLinksByRole(self, role):
        """Get a set of links by role."""
        try:
            cache = self._byrole
        except AttributeError:
            cache = OOBTree()
            for link in self:
                cache.setdefault(link.role.uri, PersistentList()).append(link)
            self._byrole = cache

        return cache.get(role.uri, [])

    def add(self, link):
        if link.__parent__ == self:
            raise ValueError("You are adding same link twice.")

        i = 1
        while "%s" % i in self._links:
            i += 1
        link.__name__ = "%s" % i
        self._links[link.__name__] = link
        link.__parent__ = self

        self._byrole.setdefault(link.role.uri, PersistentList()).append(link)

    def remove(self, link):
        if link is self._links.get(link.__name__):
            del self._links[link.__name__]
            uri = link.role.uri
            self._byrole[uri].remove(link)
            if not self._byrole[uri]:
                del self._byrole[uri]
        else:
            raise ValueError("This link does not belong to this container!")

    def clear(self):
        self._links.clear()
        self._byrole.clear()

    def __iter__(self):
        return iter(self._links.values())

    def find(self, my_role, target, role, rel_type):
        links = self.getLinksByRole(role)
        for link in links:
            if (link.rel_type == rel_type and link.target is target
                and link.my_role == my_role):
                return link
        else:
            raise ValueError(my_role, target, role, rel_type)

    def __getitem__(self, id):
        return self._links[id]

    def get(self, key, default=None):
        return self._links.get(key, default)

    def getTargetsByRole(self, role, rel_type=None):
        links = self.getLinksByRole(role)
        if rel_type is None:
            return [link.target for link in links]
        else:
            return [link.target for link in links
                    if link.rel_type == rel_type]



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2007 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool application factory useable with paste.
"""
import os

from zope.app.wsgi import WSGIPublisherApplication
from zope.app.publication.httpfactory import HTTPPublicationRequestFactory

from schooltool.app.main import SchoolToolServer


class PasteSchoolToolPublisherApplication(SchoolToolServer,
                                          WSGIPublisherApplication):

    def __init__(self, config_file, factory=HTTPPublicationRequestFactory):
        options = self.load_options(['schooltool', '-c', config_file])
        db = self.setup(options)

        super(PasteSchoolToolPublisherApplication, self).__init__(db, factory)


_st_app = None
def schooltool_app_factory(global_conf, config_file):
    global _st_app
    if _st_app:
        return _st_app
    _st_app = PasteSchoolToolPublisherApplication(
        os.path.join(global_conf['here'], config_file))
    return _st_app


def task_result_app_factory(global_conf, config_file):
    from schooltool.task.result_server import ResultServerMachinery
    machinery = ResultServerMachinery()
    machinery.configure(os.path.join(global_conf['here'], config_file))

    return machinery.app
